# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implements a single API for instrumenting all dbapi2-compliant modules
"""

from contextlib import contextmanager
from contrast_fireball import AppActivityComponentType, ArchitectureComponent, SpanType
import contrast
from contrast.agent import scope
from contrast.applies.sqli import apply_rule
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    pack_self,
    wrap_and_watermark,
)
from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")


VENDOR_TO_DB_SYSTEM = {
    "MySQL": "mysql",
    "PostgreSQL": "postgresql",
    "SQLite3": "sqlite",
}


class Dbapi2Patcher:
    """
    Do not instantiate this class; call `instrument_adapter` or subclass it.

    This class provides machinery to instrument a totally generic, PEP-249 compliant
    adapter. We only have a reference to the adapter module, and we can't make any
    assumptions about the existence of `adapter.Cursor`, since this is not guaranteed
    by the spec.

    We are only guaranteed the following:
    - the adapter has a `connection()` method, which returns an instance of Connection
    - the Connection object has a `cursor()` method, which returns an instance of Cursor
    - the Cursor has `execute()` and `executemany()` methods

    This requires a somewhat roundabout instrumentation strategy:
    - on the first call to adapter.connect(), we can instrument the Connection class
    - on the first call to Connection.cursor(), we can instrument the Cursor class
    - this lets us instrument Cursor.execute() and Cursor.executemany()
    """

    def __init__(self, adapter, vendor: str):
        self.vendor: str = vendor
        self.system: str = VENDOR_TO_DB_SYSTEM.get(vendor, "")
        self.db_name: str = ""
        self.user: str = ""
        self.host: str = ""
        self.port: str = ""

        self._adapter = adapter
        # this module name must match a policy node; we need to fake sqlalchemy
        self._adapter_name = (
            "sqlalchemy" if self.vendor == "sqlalchemy" else self._adapter.__name__
        )
        self._connect_called = False
        self._cursor_called = False

    @fail_quietly("failed to instrument database adapter")
    def instrument(self, extra_cursors):
        build_and_apply_patch(self._adapter, "connect", self._build_connect_patch)
        for cursor in extra_cursors:
            self._safe_instrument_cursor(cursor)

    @fail_quietly()
    def extract_connection_attributes(self, connection, connect_args, connect_kwargs):
        """
        This method is intended to be overridden by subclasses if necessary. It must
        either assign `self.db_name` or do nothing if the database name cannot be
        determined. "database name" is different depending on the adapter. Often this is
        a URL, but it can also be a path (sqlite3).

        Takes the arguments from a call to connect() and extracts information for
        database inventory reporting. By default only works with kwargs - we expect this
        to be by far the most common case.

        This follows the conventions laid out in the footnotes of dbapi2:
        https://peps.python.org/pep-0249/#id48

        In SQLAlchemy, see create_connect_args() for each dialect to get an idea of how
        different adapters are used. Most, if not all, use kwargs. Some adapters support
        a DSN connection string, but that should be handled by subclasses (if at all).
        """
        if not self.db_name:
            # the database name argument has several variations
            for key in ["database", "dbname", "db"]:
                if key in connect_kwargs:
                    self.db_name = str(connect_kwargs[key])
                    break
        self.user = self.user or connect_kwargs.get("user", "")
        self.host = self.host or connect_kwargs.get("host", "")
        self.port = self.port or connect_kwargs.get("port", "")

    @fail_quietly("failed to instrument database cursor class")
    def _safe_instrument_cursor(self, cursor_class):
        """
        Instruments a dbapi2-compliant database cursor class

        @param cursor_class: Reference to cursor class to be instrumented
        """
        self._instrument_cursor_method(cursor_class, "execute")
        self._instrument_cursor_method(cursor_class, "executemany")
        if hasattr(cursor_class, "executescript"):
            # non-standard, but provided by some adapters such as sqlite3
            self._instrument_cursor_method(cursor_class, "executescript")

    def _instrument_cursor_method(self, cursor, method_name):
        build_and_apply_patch(
            cursor,
            method_name,
            self._build_execute_patch,
        )

    @fail_quietly("failed to instrument database connection object")
    def _safe_instrument_connection(self, connection_instance):
        """
        Instruments a dbapi2-compliant database connection class, given an instance

        @param connection_instance: dbapi2 Connection instance
        """
        connection_class = type(connection_instance)
        build_and_apply_patch(connection_class, "cursor", self._build_cursor_patch)

    @contextmanager
    def _storage_query_span(self):
        context = contrast.REQUEST_CONTEXT.get()
        if context is None or not context.observe_enabled or scope.in_observe_scope():
            yield
            return

        with scope.observe_scope():
            if (trace := context.observability_trace) is None:
                yield
                return

            with trace.child_span(SpanType.StorageQuery) as span:
                logger.debug(
                    "entered new child span",
                    child_span=span,
                    action_type=SpanType.StorageQuery,
                )
                try:
                    yield
                finally:
                    if span:
                        db_attrs = {
                            "db.system": self.system,
                            "db.user": self.user,
                            "db.name": self.db_name,
                            "server.address": self.host,
                            "server.port": self.port,
                        }
                        db_attrs = {
                            name: value for name, value in db_attrs.items() if value
                        }
                        span.update(db_attrs)
                        logger.debug(
                            "updated child_span attributes",
                            child_span=span,
                            action_attrs=db_attrs,
                        )

    @property
    def _build_execute_patch(self):
        """
        See `_build_connect_patch`
        """

        def build_execute_patch(orig_func, _):
            def patched_method(wrapper, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection().cursor().execute*()
                """
                with self._storage_query_span():
                    return apply_rule(
                        self._adapter_name,
                        wrapper,
                        pack_self(instance, args),
                        kwargs,
                    )

            return wrap_and_watermark(orig_func, patched_method)

        return build_execute_patch

    @property
    def _build_cursor_patch(self):
        """
        See `_build_connect_patch`
        """

        def build_cursor_patch(orig_func, _):
            def cursor_patch(wrapped, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection().cursor()

                This patch will ensure that the returned Cursor object's class will have
                `execute` and `executemany` instrumented.
                """
                del instance

                cursor = wrapped(*args, **kwargs)
                if not self._cursor_called:
                    try:
                        cursor_class = type(cursor)
                        self._safe_instrument_cursor(cursor_class)
                        self._cursor_called = True
                    except Exception:
                        pass
                return cursor

            return wrap_and_watermark(orig_func, cursor_patch)

        return build_cursor_patch

    @property
    def _build_connect_patch(self):
        """
        Getter for `build_connect_patch`. We can't just make `build_connect_patch` an
        instance method because of the added `self` argument.
        """

        def build_connect_patch(orig_func, _):
            def connect_patch(wrapped, instance, args, kwargs):
                """
                Patch for dbapi_adapter.connection()

                This patch will ensure that the returned Connection object's class will
                have `cursor_patch` applied to its cursor() method.
                """
                del instance

                connection = wrapped(*args, **kwargs)
                if not self._connect_called:
                    self.extract_connection_attributes(connection, args, kwargs)
                    self._safe_instrument_connection(connection)
                    self._connect_called = True
                self._safe_send_architecture_component()
                return connection

            return wrap_and_watermark(orig_func, connect_patch)

        return build_connect_patch

    @fail_quietly()
    def _safe_send_architecture_component(self):
        """
        Report the current database information as an inventory component
        """
        from contrast.agent import agent_state

        if not (reporting_client := agent_state.module.reporting_client):
            return

        component = ArchitectureComponent(
            type=AppActivityComponentType.DB,
            url=(self.db_name or self.vendor or "default"),
            vendor=self.vendor,
        )
        reporting_client.new_inventory_components([component])


@fail_quietly("failed to instrument database adapter")
def instrument_adapter(adapter, vendor, patcher_cls=Dbapi2Patcher, extra_cursors=None):
    """
    Instrument the provided dbapi2 adapter.

    `vendor` must be a string that exactly matches a value from Teamserver's
    flowmap/technologies.json > service > one of "name".

    References to cursors to instrument explicitly can be passed in via extra_cursors.
    If we could guarantee patches were applied before any calls to `adapter.connect()`,
    we wouldn't need to directly patch extra cursors - unfortunately, this probably
    means we'd need to be a runner.
    """
    patcher_cls(adapter, vendor).instrument(extra_cursors or [])



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class EffectiveConfig(BaseTsAppMessage):
    def __init__(self):
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

        self.body = self.settings.generate_effective_config()

    @property
    def name(self):
        return "effective-config"

    @property
    def path(self):
        return (
            f"applications/{self.server_name_b64}/{self.server_path_b64}"
            f"/{self.server_type_b64}/{self.app_language_b64}/{self.app_name_b64}"
            "/effective-config"
        )

    @property
    def request_method(self):
        return requests.put

    @fail_loudly("Failed to process Effective Configuration response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast_vendor.wrapt import CallableObjectProxy

from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy.analysis import _analyze
from contrast.agent.policy import patch_manager, registry
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import (
    build_and_apply_patch,
    unregister_module_patcher,
    wrap_and_watermark,
    register_module_patcher,
)


@fail_quietly("Failed to apply assess xpath-injection")
def apply_assess(location, self, retval, args, kwargs):
    patch_policy = registry.get_policy_by_name(location)
    if patch_policy is None:
        return

    preshift = Preshift(self, args, kwargs)
    _analyze(patch_policy, preshift, self, retval, (self,) + args, kwargs)


def apply_call(class_name, orig_func, self, args, kwargs):
    result = None
    try:
        result = orig_func(*args, **kwargs)
    finally:
        location = f"lxml.etree.{class_name}.__call__"
        apply_assess(location, self, result, args, kwargs)
    return result


class ContrastXPathEvaluatorProxy(CallableObjectProxy):
    """
    Proxy class that wraps instances returned by XPathEvaluator factory

    We instrument the relevant classes directly, but since the factory is implemented
    as a C extension, the instances that it returns are the original type instead of
    our replacement. In order to cover all of our bases, we need both the replacement
    subclass and a proxied class that we return from the instrumented factory.
    """

    def __call__(__cs_self, *args, **kwargs):
        self_obj = __cs_self.__wrapped__
        orig_func = __cs_self.__wrapped__.__call__
        return apply_call(
            self_obj.__class__.__name__, orig_func, self_obj, args, kwargs
        )


def create_instrumented_xpath_element_evaluator(XPathElementEvaluator):
    """
    Generate instrumented subclass of XPathElementEvaluator

    We can't simply declare this at module level since we can't guarantee that lxml
    will be installed. We need to wait until the import hook is fired to know whether
    it's safe to make a reference to the original type.
    """

    class ContrastXPathElementEvaluator(XPathElementEvaluator):
        def __call__(__cs_self, *args, **kwargs):
            orig_func = super().__call__
            return apply_call(
                XPathElementEvaluator.__name__, orig_func, __cs_self, args, kwargs
            )

    return ContrastXPathElementEvaluator


def create_instrumented_xpath_document_evaluator(x_path_document_evaluator):
    """
    Generate instrumented subclass of XPathDocumentEvaluator

    See docstring for create_instrumented_xpath_element_evaluator above.
    """

    class ContrastXPathDocumentEvaluator(x_path_document_evaluator):
        def __call__(__cs_self, *args, **kwargs):
            orig_func = super().__call__
            return apply_call(
                x_path_document_evaluator.__name__, orig_func, __cs_self, args, kwargs
            )

    return ContrastXPathDocumentEvaluator


def create_instrumented_xpath(x_path):
    """
    Generate instrumented subclass of XPath

    We can't simply declare this at module level since we can't guarantee that lxml
    will be installed. We need to wait until the import hook is fired to know whether
    it's safe to make a reference to the original type.
    """

    class ContrastXPath(x_path):
        def __init__(__cs_self, *args, **kwargs):
            try:
                super().__init__(*args, **kwargs)
            except TypeError:
                # See PYT-2364.
                # Calling inheritance in this *bad* way fixes a TypeError raised
                # from lxml.cssselect.CSSSelector.__init__ calling
                # etree.XPath.__init__(self, ...)
                # instead of using super()
                x_path.__init__(__cs_self, *args, **kwargs)
            finally:
                apply_assess("lxml.etree.XPath.__init__", __cs_self, None, args, kwargs)

    return ContrastXPath


def create_instrumented_xml_parser(etree_xml_parser_class):
    """
    Generate a subclass of etree.XMLParser. This subclass will replace etree.XMLParser.

    We need this in order to determine if a given parser is created with the attribute
    `resolve_entities=False`. The FROMSTRING trigger action needs to be able to
    know if the parser used by `etree.fromstring` will resolve entities or not.

    There is no way to determine if a parser is safe simply by inspecting an unmodified
    XMLParser instance, because the attribute we need to check is private (Cython
    private, so we -really- can't access it).
    """

    class ContrastXMLParser(etree_xml_parser_class):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # NOTE: this is not safe by default
            self.resolve_entities = kwargs.get("resolve_entities", True)

    return ContrastXMLParser


def build_x_path_eval_patch(orig_func, _):
    def x_path_evaluator(wrapped, instance, args, kwargs):
        """
        Instrumented version of XPathEvaluator factory
        """
        del instance

        evaluator = wrapped(*args, **kwargs)
        return ContrastXPathEvaluatorProxy(evaluator)

    return wrap_and_watermark(orig_func, x_path_evaluator)


def patch_etree(etree_module):
    build_and_apply_patch(etree_module, "XPathEvaluator", build_x_path_eval_patch)

    new_xpath = create_instrumented_xpath(etree_module.XPath)
    patch_manager.patch(etree_module, "XPath", new_xpath)

    new_element_evaluator = create_instrumented_xpath_element_evaluator(
        etree_module.XPathElementEvaluator
    )
    patch_manager.patch(etree_module, "XPathElementEvaluator", new_element_evaluator)

    new_document_evaluator = create_instrumented_xpath_document_evaluator(
        etree_module.XPathDocumentEvaluator
    )
    patch_manager.patch(etree_module, "XPathDocumentEvaluator", new_document_evaluator)

    new_xml_parser = create_instrumented_xml_parser(etree_module.XMLParser)
    patch_manager.patch(etree_module, "XMLParser", new_xml_parser)


MODULE_NAME = "lxml.etree"


def register_patches():
    register_module_patcher(patch_etree, MODULE_NAME)


def reverse_patches():
    unregister_module_patcher(MODULE_NAME)
    module = sys.modules.get(MODULE_NAME)
    if not module:
        return

    patch_manager.reverse_patches_by_owner(module)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.policy import registry
from contrast.agent.assess.policy.analysis import analyze
from contrast.utils.decorators import fail_quietly


def get_propagation_node(node):
    return node.propagator_nodes[0] if node.propagator_nodes else None


@fail_quietly("Failed to analyze policy")
def analyze_policy(name, result, args, kwargs):
    context = contrast.REQUEST_CONTEXT.get()
    if context is None or not context.propagate_assess or context.stop_propagation:
        return

    context.propagated()

    with scope.contrast_scope():
        patch_policy = registry.get_policy_by_name(name)
        if not patch_policy:
            return

    analyze(patch_policy, result, args, kwargs)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.trigger_config_rule import TriggerConfigRule
from contrast.agent.assess.rules.config.session_age_rule import SessionAgeRuleMixin


class SessionTimeoutRule(TriggerConfigRule, SessionAgeRuleMixin):
    def is_violated_properties(*args, **kwargs):
        """Override for safety. In theory, this should never be called."""
        return False

    def is_violated(self, node, source, **kwargs):
        if self.count_threshold_reached():
            return False
        return SessionAgeRuleMixin.is_violated(self, source)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsServerMessage
from .effective_config import EffectiveConfig
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import now_ms


class ServerActivity(BaseTsServerMessage):
    def __init__(self):
        super().__init__()

        # NOTE: When we eventually use ServerSettings here, we'll need to use
        # `If-Modified-Since` / datetime logic. See ApplicationSettings.
        self.body = {"lastUpdate": self.since_last_update}

    @property
    def name(self):
        return "server-activity"

    @property
    def path(self):
        return "activity/server"

    @property
    def request_method(self):
        return requests.put

    @property
    def expected_response_codes(self):
        return [200, 304]

    @property
    def since_last_update(self):
        """
        Time in ms since server have been updated.
        If never updated, then it's been 0ms since then.
        """
        if self.settings.last_server_update_time_ms == 0:
            return 0
        return now_ms() - self.settings.last_server_update_time_ms

    @fail_loudly("Failed to process ServerActivity response")
    def process_response(self, response, reporting_client):
        super().process_response(response, reporting_client)

        if not self.process_response_code(response, reporting_client):
            return

        self.settings.log_effective_config()
        if (
            reporting_client is not None
            and self.settings.is_agent_config_enabled()
            # This is necessary to ensure the application exists in TeamServer
            and self.settings.config.session_id != ""
        ):
            reporting_client.add_message(EffectiveConfig())



"""
This module processes Python exceptions that relate to HTTP exceptions
by defining a set of exceptions, all subclasses of HTTPException.
Each exception, in addition to being a Python exception that can be
raised and caught, is also a WSGI application and ``webob.Response``
object.

This module defines exceptions according to RFC 2068 [1]_ : codes with
100-300 are not really errors; 400's are client errors, and 500's are
server errors.  According to the WSGI specification [2]_ , the application
can call ``start_response`` more then once only under two conditions:
(a) the response has not yet been sent, or (b) if the second and
subsequent invocations of ``start_response`` have a valid ``exc_info``
argument obtained from ``sys.exc_info()``.  The WSGI specification then
requires the server or gateway to handle the case where content has been
sent and then an exception was encountered.

Exception
  HTTPException
    HTTPOk
      * 200 - :class:`HTTPOk`
      * 201 - :class:`HTTPCreated`
      * 202 - :class:`HTTPAccepted`
      * 203 - :class:`HTTPNonAuthoritativeInformation`
      * 204 - :class:`HTTPNoContent`
      * 205 - :class:`HTTPResetContent`
      * 206 - :class:`HTTPPartialContent`
    HTTPRedirection
      * 300 - :class:`HTTPMultipleChoices`
      * 301 - :class:`HTTPMovedPermanently`
      * 302 - :class:`HTTPFound`
      * 303 - :class:`HTTPSeeOther`
      * 304 - :class:`HTTPNotModified`
      * 305 - :class:`HTTPUseProxy`
      * 307 - :class:`HTTPTemporaryRedirect`
      * 308 - :class:`HTTPPermanentRedirect`
    HTTPError
      HTTPClientError
        * 400 - :class:`HTTPBadRequest`
        * 401 - :class:`HTTPUnauthorized`
        * 402 - :class:`HTTPPaymentRequired`
        * 403 - :class:`HTTPForbidden`
        * 404 - :class:`HTTPNotFound`
        * 405 - :class:`HTTPMethodNotAllowed`
        * 406 - :class:`HTTPNotAcceptable`
        * 407 - :class:`HTTPProxyAuthenticationRequired`
        * 408 - :class:`HTTPRequestTimeout`
        * 409 - :class:`HTTPConflict`
        * 410 - :class:`HTTPGone`
        * 411 - :class:`HTTPLengthRequired`
        * 412 - :class:`HTTPPreconditionFailed`
        * 413 - :class:`HTTPRequestEntityTooLarge`
        * 414 - :class:`HTTPRequestURITooLong`
        * 415 - :class:`HTTPUnsupportedMediaType`
        * 416 - :class:`HTTPRequestRangeNotSatisfiable`
        * 417 - :class:`HTTPExpectationFailed`
        * 422 - :class:`HTTPUnprocessableEntity`
        * 423 - :class:`HTTPLocked`
        * 424 - :class:`HTTPFailedDependency`
        * 428 - :class:`HTTPPreconditionRequired`
        * 429 - :class:`HTTPTooManyRequests`
        * 431 - :class:`HTTPRequestHeaderFieldsTooLarge`
        * 451 - :class:`HTTPUnavailableForLegalReasons`
      HTTPServerError
        * 500 - :class:`HTTPInternalServerError`
        * 501 - :class:`HTTPNotImplemented`
        * 502 - :class:`HTTPBadGateway`
        * 503 - :class:`HTTPServiceUnavailable`
        * 504 - :class:`HTTPGatewayTimeout`
        * 505 - :class:`HTTPVersionNotSupported`
        * 511 - :class:`HTTPNetworkAuthenticationRequired`

Usage notes
-----------

The HTTPException class is complicated by 4 factors:

  1. The content given to the exception may either be plain-text or
     as html-text.

  2. The template may want to have string-substitutions taken from
     the current ``environ`` or values from incoming headers. This
     is especially troublesome due to case sensitivity.

  3. The final output may either be text/plain or text/html
     mime-type as requested by the client application.

  4. Each exception has a default explanation, but those who
     raise exceptions may want to provide additional detail.

Subclass attributes and call parameters are designed to provide an easier path
through the complications.

Attributes:

   ``code``
       the HTTP status code for the exception

   ``title``
       remainder of the status line (stuff after the code)

   ``explanation``
       a plain-text explanation of the error message that is
       not subject to environment or header substitutions;
       it is accessible in the template via %(explanation)s

   ``detail``
       a plain-text message customization that is not subject
       to environment or header substitutions; accessible in
       the template via %(detail)s

   ``body_template``
       a content fragment (in HTML) used for environment and
       header substitution; the default template includes both
       the explanation and further detail provided in the
       message

Parameters:

   ``detail``
     a plain-text override of the default ``detail``

   ``headers``
     a list of (k,v) header pairs

   ``comment``
     a plain-text additional information which is
     usually stripped/hidden for end-users

   ``body_template``
     a string.Template object containing a content fragment in HTML
     that frames the explanation and further detail

To override the template (which is HTML content) or the plain-text
explanation, one must subclass the given exception; or customize it
after it has been created.  This particular breakdown of a message
into explanation, detail and template allows both the creation of
plain-text and html messages for various clients as well as
error-free substitution of environment variables and headers.


The subclasses of :class:`~_HTTPMove`
(:class:`~HTTPMultipleChoices`, :class:`~HTTPMovedPermanently`,
:class:`~HTTPFound`, :class:`~HTTPSeeOther`, :class:`~HTTPUseProxy` and
:class:`~HTTPTemporaryRedirect`) are redirections that require a ``Location``
field. Reflecting this, these subclasses have two additional keyword arguments:
``location`` and ``add_slash``.

Parameters:

    ``location``
      to set the location immediately

    ``add_slash``
      set to True to redirect to the same URL as the request, except with a
      ``/`` appended

Relative URLs in the location will be resolved to absolute.

References:

.. [1] https://www.python.org/dev/peps/pep-0333/#error-handling
.. [2] https://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10.5


"""

import json
from string import Template
import re
import sys

from contrast_vendor.webob.acceptparse import create_accept_header
from contrast_vendor.webob.compat import (
    class_types,
    text_,
    text_type,
    urlparse,
    )
from contrast_vendor.webob.request import Request
from contrast_vendor.webob.response import Response
from contrast_vendor.webob.util import html_escape

tag_re = re.compile(r'<.*?>', re.S)
br_re = re.compile(r'<br.*?>', re.I | re.S)
comment_re = re.compile(r'<!--|-->')

class _lazified(object):
    def __init__(self, func, value):
        self.func = func
        self.value = value

    def __str__(self):
        return self.func(self.value)

def lazify(func):
    def wrapper(value):
        return _lazified(func, value)
    return wrapper

def no_escape(value):
    if value is None:
        return ''
    if not isinstance(value, text_type):
        if hasattr(value, '__unicode__'):
            value = value.__unicode__()
        if isinstance(value, bytes):
            value = text_(value, 'utf-8')
        else:
            value = text_type(value)
    return value

def strip_tags(value):
    value = value.replace('\n', ' ')
    value = value.replace('\r', '')
    value = br_re.sub('\n', value)
    value = comment_re.sub('', value)
    value = tag_re.sub('', value)
    return value

class HTTPException(Exception):
    def __init__(self, message, wsgi_response):
        Exception.__init__(self, message)
        self.wsgi_response = wsgi_response

    def __call__(self, environ, start_response):
        return self.wsgi_response(environ, start_response)

class WSGIHTTPException(Response, HTTPException):

    ## You should set in subclasses:
    # code = 200
    # title = 'OK'
    # explanation = 'why this happens'
    # body_template_obj = Template('response template')
    code = 500
    title = 'Internal Server Error'
    explanation = ''
    body_template_obj = Template('''\
${explanation}<br /><br />
${detail}
${html_comment}
''')

    plain_template_obj = Template('''\
${status}

${body}''')

    html_template_obj = Template('''\
<html>
 <head>
  <title>${status}</title>
 </head>
 <body>
  <h1>${status}</h1>
  ${body}
 </body>
</html>''')

    ## Set this to True for responses that should have no request body
    empty_body = False

    def __init__(self, detail=None, headers=None, comment=None,
                 body_template=None, json_formatter=None, **kw):
        Response.__init__(self,
                          status='%s %s' % (self.code, self.title),
                          **kw)
        Exception.__init__(self, detail)
        if headers:
            self.headers.extend(headers)
        self.detail = detail
        self.comment = comment
        if body_template is not None:
            self.body_template = body_template
            self.body_template_obj = Template(body_template)
        if self.empty_body:
            del self.content_type
            del self.content_length
        if json_formatter is not None:
            self.json_formatter = json_formatter

    def __str__(self):
        return self.detail or self.explanation

    def _make_body(self, environ, escape):
        escape = lazify(escape)
        args = {
            'explanation': escape(self.explanation),
            'detail': escape(self.detail or ''),
            'comment': escape(self.comment or ''),
            }
        if self.comment:
            args['html_comment'] = '<!-- %s -->' % escape(self.comment)
        else:
            args['html_comment'] = ''
        if WSGIHTTPException.body_template_obj is not self.body_template_obj:
            # Custom template; add headers to args
            for k, v in environ.items():
                args[k] = escape(v)
            for k, v in self.headers.items():
                args[k.lower()] = escape(v)
        t_obj = self.body_template_obj
        return t_obj.safe_substitute(args)

    def plain_body(self, environ):
        body = self._make_body(environ, no_escape)
        body = strip_tags(body)
        return self.plain_template_obj.substitute(status=self.status,
                                                  title=self.title,
                                                  body=body)

    def html_body(self, environ):
        body = self._make_body(environ, html_escape)
        return self.html_template_obj.substitute(status=self.status,
                                                 body=body)

    def json_formatter(self, body, status, title, environ):
        return {'message': body,
                'code': status,
                'title': title}

    def json_body(self, environ):
        body = self._make_body(environ, no_escape)
        jsonbody = self.json_formatter(body=body, status=self.status,
                                       title=self.title, environ=environ)
        return json.dumps(jsonbody)

    def generate_response(self, environ, start_response):
        if self.content_length is not None:
            del self.content_length
        headerlist = list(self.headerlist)
        accept_value = environ.get('HTTP_ACCEPT', '')
        accept_header = create_accept_header(header_value=accept_value)
        acceptable_offers = accept_header.acceptable_offers(
            offers=['text/html', 'application/json'],
        )
        match = acceptable_offers[0][0] if acceptable_offers else None

        if match == 'text/html':
            content_type = 'text/html'
            body = self.html_body(environ)
        elif match == 'application/json':
            content_type = 'application/json'
            body = self.json_body(environ)
        else:
            content_type = 'text/plain'
            body = self.plain_body(environ)
        resp = Response(body,
                        status=self.status,
                        headerlist=headerlist,
                        content_type=content_type,
                        )
        resp.content_type = content_type
        return resp(environ, start_response)

    def __call__(self, environ, start_response):
        is_head = environ['REQUEST_METHOD'] == 'HEAD'
        if self.has_body or self.empty_body or is_head:
            app_iter = Response.__call__(self, environ, start_response)
        else:
            app_iter = self.generate_response(environ, start_response)
        if is_head:
            app_iter = []
        return app_iter

    @property
    def wsgi_response(self):
        return self



class HTTPError(WSGIHTTPException):
    """
    base class for status codes in the 400's and 500's

    This is an exception which indicates that an error has occurred,
    and that any work in progress should not be committed.  These are
    typically results in the 400's and 500's.
    """

class HTTPRedirection(WSGIHTTPException):
    """
    base class for 300's status code (redirections)

    This is an abstract base class for 3xx redirection.  It indicates
    that further action needs to be taken by the user agent in order
    to fulfill the request.  It does not necessarly signal an error
    condition.
    """

class HTTPOk(WSGIHTTPException):
    """
    Base class for the 200's status code (successful responses)

    code: 200, title: OK
    """
    code = 200
    title = 'OK'

############################################################
## 2xx success
############################################################

class HTTPCreated(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that request has been fulfilled and resulted in a new
    resource being created.

    code: 201, title: Created
    """
    code = 201
    title = 'Created'

class HTTPAccepted(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that the request has been accepted for processing, but the
    processing has not been completed.

    code: 202, title: Accepted
    """
    code = 202
    title = 'Accepted'
    explanation = 'The request is accepted for processing.'

class HTTPNonAuthoritativeInformation(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that the returned metainformation in the entity-header is
    not the definitive set as available from the origin server, but is
    gathered from a local or a third-party copy.

    code: 203, title: Non-Authoritative Information
    """
    code = 203
    title = 'Non-Authoritative Information'

class HTTPNoContent(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that the server has fulfilled the request but does
    not need to return an entity-body, and might want to return updated
    metainformation.

    code: 204, title: No Content
    """
    code = 204
    title = 'No Content'
    empty_body = True

class HTTPResetContent(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that the the server has fulfilled the request and
    the user agent SHOULD reset the document view which caused the
    request to be sent.

    code: 205, title: Reset Content
    """
    code = 205
    title = 'Reset Content'
    empty_body = True

class HTTPPartialContent(HTTPOk):
    """
    subclass of :class:`~HTTPOk`

    This indicates that the server has fulfilled the partial GET
    request for the resource.

    code: 206, title: Partial Content
    """
    code = 206
    title = 'Partial Content'

############################################################
## 3xx redirection
############################################################

class _HTTPMove(HTTPRedirection):
    """
    redirections which require a Location field

    Since a 'Location' header is a required attribute of 301, 302, 303,
    305, 307 and 308 (but not 304), this base class provides the mechanics to
    make this easy.

    You can provide a location keyword argument to set the location
    immediately.  You may also give ``add_slash=True`` if you want to
    redirect to the same URL as the request, except with a ``/`` added
    to the end.

    Relative URLs in the location will be resolved to absolute.
    """
    explanation = 'The resource has been moved to'
    body_template_obj = Template('''\
${explanation} <a href="${location}">${location}</a>;
you should be redirected automatically.
${detail}
${html_comment}''')

    def __init__(self, detail=None, headers=None, comment=None,
                 body_template=None, location=None, add_slash=False):
        super(_HTTPMove, self).__init__(
            detail=detail, headers=headers, comment=comment,
            body_template=body_template)
        if location is not None:
            if '\n' in location or '\r' in location:
                raise ValueError('Control characters are not allowed in location')

            self.location = location
            if add_slash:
                raise TypeError(
                    "You can only provide one of the arguments location "
                    "and add_slash")
        self.add_slash = add_slash

    def __call__(self, environ, start_response):
        req = Request(environ)
        if self.add_slash:
            url = req.path_url
            url += '/'
            if req.environ.get('QUERY_STRING'):
                url += '?' + req.environ['QUERY_STRING']
            self.location = url
        self.location = urlparse.urljoin(req.path_url, self.location)
        return super(_HTTPMove, self).__call__(
            environ, start_response)

class HTTPMultipleChoices(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource corresponds to any one
    of a set of representations, each with its own specific location,
    and agent-driven negotiation information is being provided so that
    the user can select a preferred representation and redirect its
    request to that location.

    code: 300, title: Multiple Choices
    """
    code = 300
    title = 'Multiple Choices'

class HTTPMovedPermanently(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource has been assigned a new
    permanent URI and any future references to this resource SHOULD use
    one of the returned URIs.

    code: 301, title: Moved Permanently
    """
    code = 301
    title = 'Moved Permanently'

class HTTPFound(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource resides temporarily under
    a different URI.

    code: 302, title: Found
    """
    code = 302
    title = 'Found'
    explanation = 'The resource was found at'

# This one is safe after a POST (the redirected location will be
# retrieved with GET):
class HTTPSeeOther(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the response to the request can be found under
    a different URI and SHOULD be retrieved using a GET method on that
    resource.

    code: 303, title: See Other
    """
    code = 303
    title = 'See Other'

class HTTPNotModified(HTTPRedirection):
    """
    subclass of :class:`~HTTPRedirection`

    This indicates that if the client has performed a conditional GET
    request and access is allowed, but the document has not been
    modified, the server SHOULD respond with this status code.

    code: 304, title: Not Modified
    """
    # TODO: this should include a date or etag header
    code = 304
    title = 'Not Modified'
    empty_body = True

class HTTPUseProxy(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource MUST be accessed through
    the proxy given by the Location field.

    code: 305, title: Use Proxy
    """
    # Not a move, but looks a little like one
    code = 305
    title = 'Use Proxy'
    explanation = (
        'The resource must be accessed through a proxy located at')

class HTTPTemporaryRedirect(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource resides temporarily
    under a different URI.

    code: 307, title: Temporary Redirect
    """
    code = 307
    title = 'Temporary Redirect'

class HTTPPermanentRedirect(_HTTPMove):
    """
    subclass of :class:`~_HTTPMove`

    This indicates that the requested resource resides permanently
    under a different URI.

    code: 308, title: Permanent Redirect
    """
    code = 308
    title = 'Permanent Redirect'


############################################################
## 4xx client error
############################################################

class HTTPClientError(HTTPError):
    """
    base class for the 400's, where the client is in error

    This is an error condition in which the client is presumed to be
    in-error.  This is an expected problem, and thus is not considered
    a bug.  A server-side traceback is not warranted.  Unless specialized,
    this is a '400 Bad Request'

    code: 400, title: Bad Request
    """
    code = 400
    title = 'Bad Request'
    explanation = ('The server could not comply with the request since\r\n'
                   'it is either malformed or otherwise incorrect.\r\n')

class HTTPBadRequest(HTTPClientError):
    pass

class HTTPUnauthorized(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the request requires user authentication.

    code: 401, title: Unauthorized
    """
    code = 401
    title = 'Unauthorized'
    explanation = (
        'This server could not verify that you are authorized to\r\n'
        'access the document you requested.  Either you supplied the\r\n'
        'wrong credentials (e.g., bad password), or your browser\r\n'
        'does not understand how to supply the credentials required.\r\n')

class HTTPPaymentRequired(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    code: 402, title: Payment Required
    """
    code = 402
    title = 'Payment Required'
    explanation = ('Access was denied for financial reasons.')

class HTTPForbidden(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server understood the request, but is
    refusing to fulfill it.

    code: 403, title: Forbidden
    """
    code = 403
    title = 'Forbidden'
    explanation = ('Access was denied to this resource.')

class HTTPNotFound(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server did not find anything matching the
    Request-URI.

    code: 404, title: Not Found
    """
    code = 404
    title = 'Not Found'
    explanation = ('The resource could not be found.')

class HTTPMethodNotAllowed(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the method specified in the Request-Line is
    not allowed for the resource identified by the Request-URI.

    code: 405, title: Method Not Allowed
    """
    code = 405
    title = 'Method Not Allowed'
    # override template since we need an environment variable
    body_template_obj = Template('''\
The method ${REQUEST_METHOD} is not allowed for this resource. <br /><br />
${detail}''')

class HTTPNotAcceptable(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates the resource identified by the request is only
    capable of generating response entities which have content
    characteristics not acceptable according to the accept headers
    sent in the request.

    code: 406, title: Not Acceptable
    """
    code = 406
    title = 'Not Acceptable'
    # override template since we need an environment variable
    body_template_obj = Template('''\
The resource could not be generated that was acceptable to your browser
(content of type ${HTTP_ACCEPT}. <br /><br />
${detail}''')

class HTTPProxyAuthenticationRequired(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This is similar to 401, but indicates that the client must first
    authenticate itself with the proxy.

    code: 407, title: Proxy Authentication Required
    """
    code = 407
    title = 'Proxy Authentication Required'
    explanation = ('Authentication with a local proxy is needed.')

class HTTPRequestTimeout(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the client did not produce a request within
    the time that the server was prepared to wait.

    code: 408, title: Request Timeout
    """
    code = 408
    title = 'Request Timeout'
    explanation = ('The server has waited too long for the request to '
                   'be sent by the client.')

class HTTPConflict(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the request could not be completed due to a
    conflict with the current state of the resource.

    code: 409, title: Conflict
    """
    code = 409
    title = 'Conflict'
    explanation = ('There was a conflict when trying to complete '
                   'your request.')

class HTTPGone(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the requested resource is no longer available
    at the server and no forwarding address is known.

    code: 410, title: Gone
    """
    code = 410
    title = 'Gone'
    explanation = ('This resource is no longer available.  No forwarding '
                   'address is given.')

class HTTPLengthRequired(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the the server refuses to accept the request
    without a defined Content-Length.

    code: 411, title: Length Required
    """
    code = 411
    title = 'Length Required'
    explanation = ('Content-Length header required.')

class HTTPPreconditionFailed(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the precondition given in one or more of the
    request-header fields evaluated to false when it was tested on the
    server.

    code: 412, title: Precondition Failed
    """
    code = 412
    title = 'Precondition Failed'
    explanation = ('Request precondition failed.')

class HTTPRequestEntityTooLarge(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is refusing to process a request
    because the request entity is larger than the server is willing or
    able to process.

    code: 413, title: Request Entity Too Large
    """
    code = 413
    title = 'Request Entity Too Large'
    explanation = ('The body of your request was too large for this server.')

class HTTPRequestURITooLong(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is refusing to service the request
    because the Request-URI is longer than the server is willing to
    interpret.

    code: 414, title: Request-URI Too Long
    """
    code = 414
    title = 'Request-URI Too Long'
    explanation = ('The request URI was too long for this server.')

class HTTPUnsupportedMediaType(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is refusing to service the request
    because the entity of the request is in a format not supported by
    the requested resource for the requested method.

    code: 415, title: Unsupported Media Type
    """
    code = 415
    title = 'Unsupported Media Type'
    # override template since we need an environment variable
    body_template_obj = Template('''\
The request media type ${CONTENT_TYPE} is not supported by this server.
<br /><br />
${detail}''')

class HTTPRequestRangeNotSatisfiable(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    The server SHOULD return a response with this status code if a
    request included a Range request-header field, and none of the
    range-specifier values in this field overlap the current extent
    of the selected resource, and the request did not include an
    If-Range request-header field.

    code: 416, title: Request Range Not Satisfiable
    """
    code = 416
    title = 'Request Range Not Satisfiable'
    explanation = ('The Range requested is not available.')

class HTTPExpectationFailed(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indidcates that the expectation given in an Expect
    request-header field could not be met by this server.

    code: 417, title: Expectation Failed
    """
    code = 417
    title = 'Expectation Failed'
    explanation = ('Expectation failed.')

class HTTPUnprocessableEntity(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is unable to process the contained
    instructions.

    code: 422, title: Unprocessable Entity
    """
    ## Note: from WebDAV
    code = 422
    title = 'Unprocessable Entity'
    explanation = 'Unable to process the contained instructions'

class HTTPLocked(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the resource is locked.

    code: 423, title: Locked
    """
    ## Note: from WebDAV
    code = 423
    title = 'Locked'
    explanation = ('The resource is locked')

class HTTPFailedDependency(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the method could not be performed because the
    requested action depended on another action and that action failed.

    code: 424, title: Failed Dependency
    """
    ## Note: from WebDAV
    code = 424
    title = 'Failed Dependency'
    explanation = (
        'The method could not be performed because the requested '
        'action dependended on another action and that action failed')

class HTTPPreconditionRequired(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the origin server requires the request to be
    conditional.  From RFC 6585, "Additional HTTP Status Codes".

    code: 428, title: Precondition Required
    """
    code = 428
    title = 'Precondition Required'
    explanation = ('This request is required to be conditional')

class HTTPTooManyRequests(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the client has sent too many requests in a
    given amount of time.  Useful for rate limiting.

    From RFC 6585, "Additional HTTP Status Codes".

    code: 429, title: Too Many Requests
    """
    code = 429
    title = 'Too Many Requests'
    explanation = (
        'The client has sent too many requests in a given amount of time')

class HTTPRequestHeaderFieldsTooLarge(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is unwilling to process the request
    because its header fields are too large. The request may be resubmitted
    after reducing the size of the request header fields.

    From RFC 6585, "Additional HTTP Status Codes".

    code: 431, title: Request Header Fields Too Large
    """
    code = 431
    title = 'Request Header Fields Too Large'
    explanation = (
        'The request header fields were too large')

class HTTPUnavailableForLegalReasons(HTTPClientError):
    """
    subclass of :class:`~HTTPClientError`

    This indicates that the server is unable to process the request
    because of legal reasons, e.g. censorship or government-mandated
    blocked access.

    From the draft "A New HTTP Status Code for Legally-restricted Resources"
    by Tim Bray:

    https://tools.ietf.org/html/draft-tbray-http-legally-restricted-status-00

    code: 451, title: Unavailable For Legal Reasons
    """
    code = 451
    title = 'Unavailable For Legal Reasons'
    explanation = ('The resource is not available due to legal reasons.')

############################################################
## 5xx Server Error
############################################################
#  Response status codes beginning with the digit "5" indicate cases in
#  which the server is aware that it has erred or is incapable of
#  performing the request. Except when responding to a HEAD request, the
#  server SHOULD include an entity containing an explanation of the error
#  situation, and whether it is a temporary or permanent condition. User
#  agents SHOULD display any included entity to the user. These response
#  codes are applicable to any request method.

class HTTPServerError(HTTPError):
    """
    base class for the 500's, where the server is in-error

    This is an error condition in which the server is presumed to be
    in-error.  This is usually unexpected, and thus requires a traceback;
    ideally, opening a support ticket for the customer. Unless specialized,
    this is a '500 Internal Server Error'
    """
    code = 500
    title = 'Internal Server Error'
    explanation = (
      'The server has either erred or is incapable of performing\r\n'
      'the requested operation.\r\n')

class HTTPInternalServerError(HTTPServerError):
    pass

class HTTPNotImplemented(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server does not support the functionality
    required to fulfill the request.

    code: 501, title: Not Implemented
    """
    code = 501
    title = 'Not Implemented'
    body_template_obj = Template('''
The request method ${REQUEST_METHOD} is not implemented for this server. <br /><br />
${detail}''')

class HTTPBadGateway(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server, while acting as a gateway or proxy,
    received an invalid response from the upstream server it accessed
    in attempting to fulfill the request.

    code: 502, title: Bad Gateway
    """
    code = 502
    title = 'Bad Gateway'
    explanation = ('Bad gateway.')

class HTTPServiceUnavailable(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server is currently unable to handle the
    request due to a temporary overloading or maintenance of the server.

    code: 503, title: Service Unavailable
    """
    code = 503
    title = 'Service Unavailable'
    explanation = ('The server is currently unavailable. '
                   'Please try again at a later time.')

class HTTPGatewayTimeout(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server, while acting as a gateway or proxy,
    did not receive a timely response from the upstream server specified
    by the URI (e.g. HTTP, FTP, LDAP) or some other auxiliary server
    (e.g. DNS) it needed to access in attempting to complete the request.

    code: 504, title: Gateway Timeout
    """
    code = 504
    title = 'Gateway Timeout'
    explanation = ('The gateway has timed out.')

class HTTPVersionNotSupported(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server does not support, or refuses to
    support, the HTTP protocol version that was used in the request
    message.

    code: 505, title: HTTP Version Not Supported
    """
    code = 505
    title = 'HTTP Version Not Supported'
    explanation = ('The HTTP version is not supported.')

class HTTPInsufficientStorage(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the server does not have enough space to save
    the resource.

    code: 507, title: Insufficient Storage
    """
    code = 507
    title = 'Insufficient Storage'
    explanation = ('There was not enough space to save the resource')

class HTTPNetworkAuthenticationRequired(HTTPServerError):
    """
    subclass of :class:`~HTTPServerError`

    This indicates that the client needs to authenticate to gain
    network access.  From RFC 6585, "Additional HTTP Status Codes".

    code: 511, title: Network Authentication Required
    """
    code = 511
    title = 'Network Authentication Required'
    explanation = ('Network authentication is required')

class HTTPExceptionMiddleware(object):
    """
    Middleware that catches exceptions in the sub-application.  This
    does not catch exceptions in the app_iter; only during the initial
    calling of the application.

    This should be put *very close* to applications that might raise
    these exceptions.  This should not be applied globally; letting
    *expected* exceptions raise through the WSGI stack is dangerous.
    """

    def __init__(self, application):
        self.application = application
    def __call__(self, environ, start_response):
        try:
            return self.application(environ, start_response)
        except HTTPException:
            parent_exc_info = sys.exc_info()
            def repl_start_response(status, headers, exc_info=None):
                if exc_info is None:
                    exc_info = parent_exc_info
                return start_response(status, headers, exc_info)
            return parent_exc_info[1](environ, repl_start_response)

try:
    from paste import httpexceptions
except ImportError:   # pragma: no cover
    # Without Paste we don't need to do this fixup
    pass
else: # pragma: no cover
    for name in dir(httpexceptions):
        obj = globals().get(name)
        if (obj and isinstance(obj, type) and issubclass(obj, HTTPException)
            and obj is not HTTPException
            and obj is not WSGIHTTPException):
            obj.__bases__ = obj.__bases__ + (getattr(httpexceptions, name),)
    del name, obj, httpexceptions

__all__ = ['HTTPExceptionMiddleware', 'status_map']
status_map={}
for name, value in list(globals().items()):
    if (isinstance(value, (type, class_types)) and
        issubclass(value, HTTPException)
        and not name.startswith('_')):
        __all__.append(name)
        if all((
            getattr(value, 'code', None),
            value not in (HTTPRedirection, HTTPClientError, HTTPServerError),
            issubclass(
                value,
                (HTTPOk, HTTPRedirection, HTTPClientError, HTTPServerError)
            )
        )):
            status_map[value.code]=value
        if hasattr(value, 'explanation'):
            value.explanation = ' '.join(value.explanation.strip().split())
del name, value



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from functools import cached_property, lru_cache

import contrast
from contrast.agent import agent_state, request_state, scope as scope_
from contrast.agent.framework import UNKNOWN_FRAMEWORK
from contrast.agent.policy.trigger_node import TriggerNode
from contrast.agent.middlewares.base_middleware import (
    BaseMiddleware,
)
from contrast.agent.asgi import (
    track_scope_sources,
    ASGIRequest,
    ASGIResponse,
)
from contrast.agent.request_context import RequestContext
from contrast.utils.safe_import import safe_import_list
from contrast.utils.assess.duck_utils import safe_getattr_list

from contrast_vendor import structlog as logging


logger = logging.getLogger("contrast")

DEFAULT_ASGI_NAME = "asgi_app"


@lru_cache(maxsize=1)
def _get_quart_types():
    return tuple(safe_import_list("quart.Quart"))


class ASGIMiddleware(BaseMiddleware):
    @scope_.contrast_scope()
    def __init__(
        self,
        app,
        app_name=None,
        original_app=None,
        *,
        framework_name: str = UNKNOWN_FRAMEWORK.name,
    ):
        """
        Contrast middleware - ASGI 3.0

        The `app` argument here must be named "app" exactly to support FastAPI's
        middleware system, which uses the kwarg when calling this method.

        We need to keep the `original_app` kwarg for now to prevent a breaking API
        change.
        """
        del original_app

        # TODO: PYT-2852 Revisit application name detection
        self.app_name = (
            app_name
            if app_name is not None
            else safe_getattr_list(
                app,
                [
                    "__name__",
                    "name",
                    "title",
                ],
                DEFAULT_ASGI_NAME,
            )
        )
        self.framework_name = framework_name
        agent_state.set_detected_framework(self.framework_name)

        super().__init__()

        if not agent_state.in_automatic_middleware() and isinstance(
            app, _get_quart_types()
        ):
            # we need this to prevent a breaking API change when wrapping Quart apps
            app = app.asgi_app

        self.asgi_app = app

    async def __call__(self, scope, receive, send) -> None:
        if request_state.get_request_id() is not None:
            # This can happen if a single app is wrapped by multiple instances of the
            # middleware (usually caused by automatic instrumentation)
            logger.debug("Detected preexisting request_id - passing through")
            await self.asgi_app(scope, receive, send)
            return

        request_path = scope.get("path", "")
        request_method = scope.get("method", "")

        if scope.get("type") != "http":
            logger.debug("Detected non-http request - cannot analyze", scope=scope)
            await self.call_without_agent_async(scope, receive, send)
            return

        with self.request_context_stack(
            request_method, request_path
        ) as request_context_stack:
            request = ASGIRequest(scope, receive)
            environ = await request.to_wsgi_environ()

            context = self.should_analyze_request(environ)
            if context:
                self.enter_agent_call_contexts(request_context_stack, context)
                await self.call_with_agent(context, request, send)
                return

            await self.call_without_agent_async(scope, request.contrast__receive, send)

    async def call_with_agent(self, context: RequestContext, request, send) -> None:
        with scope_.contrast_scope():
            track_scope_sources(context, request.scope)

            try:
                self.prefilter(context)

                response = ASGIResponse(send)
                with scope_.pop_contrast_scope():
                    try:
                        await self.asgi_app(
                            request.scope,
                            request.contrast__receive,
                            response.contrast__send,
                        )
                        context.extract_response(response)
                    except Exception as exc:
                        context.response_exception = exc
                        raise

                self.postfilter(context)
                self.check_for_blocked(context)

            finally:
                self.handle_ensure(context, context.request)
                if context.assess_enabled:
                    contrast.STRING_TRACKER.ageoff()

    async def call_without_agent_async(self, scope, receive, send) -> None:
        with scope_.contrast_scope():
            await self.asgi_app(scope, receive, send)

    @cached_property
    def trigger_node(self):
        """
        trigger node used by reflected xss postfilter rule
        """
        method_name = self.app_name

        module, class_name, args, instance_method = self._process_trigger_handler(
            self.asgi_app
        )

        return (
            TriggerNode(module, class_name, instance_method, method_name, "RETURN"),
            args,
        )

    @cached_property
    def name(self):
        return "asgi"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contextlib
import sys
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)

MYSQL_CONNECTOR = "mysql.connector"
VENDOR = "MySQL"


def instrument_mysql_connector(mysql_connector):
    extra_cursors = [
        mysql_connector.cursor.MySQLCursor,
    ]

    # CursorBase is a base class for MySQLCursor prior to 9.0.0,
    # when it was removed.
    if cb := getattr(mysql_connector.cursor, "CursorBase", None):
        extra_cursors.append(cb)

    with contextlib.suppress(AttributeError):
        # The C extension is technically optional
        extra_cursors.append(mysql_connector.cursor_cext.CMySQLCursor)

    dbapi2.instrument_adapter(
        mysql_connector,
        VENDOR,
        dbapi2.Dbapi2Patcher,
        extra_cursors=extra_cursors,
    )


def register_patches():
    register_module_patcher(instrument_mysql_connector, MYSQL_CONNECTOR)


def reverse_patches():
    unregister_module_patcher(MYSQL_CONNECTOR)
    if mysql_connector := sys.modules.get(MYSQL_CONNECTOR):
        patch_manager.reverse_patches_by_owner(mysql_connector)
        patch_manager.reverse_patches_by_owner(
            mysql_connector.connection_cext.CMySQLConnection
        )
        if cb := getattr(mysql_connector.cursor, "CursorBase", None):
            patch_manager.reverse_patches_by_owner(cb)
        patch_manager.reverse_patches_by_owner(mysql_connector.cursor.MySQLCursor)
        patch_manager.reverse_patches_by_owner(mysql_connector.cursor_cext.CMySQLCursor)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from collections.abc import Mapping
from typing import Callable

from contrast_fireball import OtelAttributes, SpanType

from contrast.agent.policy.handlers import EventDict


def django_authn_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    mechanism_param_locations = event_dict.get("mechanisms", [])
    if not mechanism_param_locations:
        raise ValueError("Event must specify 'mechanisms' parameter location.")

    def observe_span_attrs(args: Mapping[str, object], result: object):
        credentials = args["credentials"]
        mechanism = next(
            (m for m in mechanism_param_locations if credentials.get(m)), None
        )
        return (
            {"contrast.authentication.mechanism": str(mechanism)} if mechanism else {}
        )

    return SpanType.AuthenticationRequest, observe_span_attrs, None


def django_session_authn_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    mechanism = str(event_dict.get("mechanism", ""))
    if not mechanism:
        raise ValueError("Event must specify 'mechanism'.")

    def observe_span_attrs(args: Mapping[str, object], result: object):
        return {"contrast.authentication.mechanism": mechanism}

    return SpanType.AuthenticationRequest, observe_span_attrs, None


def starlette_authn_span_attrs_builder(
    event_dict: EventDict,
) -> tuple[SpanType, Callable[[Mapping[str, object], object], OtelAttributes], None]:
    def observe_span_attrs(args: Mapping[str, object], result: object):
        # Starlette authentication requires the app developer to implement their own
        # AuthenticationBackend, so we have no way of knowing the mechanism used.
        # Still, this will signal that authentication was attempted.
        return {}

    return SpanType.AuthenticationRequest, observe_span_attrs, None



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsMessage


class AgentConnection(BaseTsMessage):
    def __init__(self):
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

    @property
    def name(self):
        return "connection"

    @property
    def path(self):
        return "agents/connection"

    @property
    def request_method(self):
        return requests.get

    @property
    def expected_response_codes(self):
        return [204]

    def process_response(self, response, reporting_client):
        """This endpoint doesn't have a response body, so there's nothing to process."""



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from contrast.agent.assess.tag import Tag


ELLIPSIS = "..."
TRUNCATION_FRAGMENT_LENGTH = 50
TRUNCATION_LENGTH = TRUNCATION_FRAGMENT_LENGTH * 2 + len(ELLIPSIS)


def truncate(str_representation, length_factor=1):
    """
    Truncate + return the provided string
    """
    fragment_length = TRUNCATION_FRAGMENT_LENGTH * length_factor
    if len(str_representation) <= 2 * fragment_length + len(ELLIPSIS):
        return str_representation
    return "".join(
        [
            str_representation[:fragment_length],
            ELLIPSIS,
            str_representation[-fragment_length:],
        ]
    )


def truncate_tainted_string(
    str_representation: str, tags: list[Tag] | None
) -> tuple[str, list[Tag]]:
    """
    Truncate + return the provided string and update the provided event's taint_ranges
    accordingly.

    This function assumes:
    - event.taint_ranges has already been calculated
    - each range in event.taint_ranges follows the format "{}:{}"
    - str_representation is the string representation of the tainted object associated
      with the given event, i.e. event.taint_ranges describes str_representation
    """
    if len(str_representation) <= TRUNCATION_LENGTH:
        return str_representation, tags or []
    if not tags:
        return truncate(str_representation), tags or []

    return _truncate_tainted_string(str_representation, tags)


def _truncate_tainted_string(
    str_representation: str, merged_tags: list[Tag]
) -> tuple[str, list[Tag]]:
    """
    The core logic of tainted string truncation. Given a string and a list of simplified
    and merged Tags, truncate the string approximately according to this specification:

    https://github.com/Contrast-Security-Inc/assess-specifications/blob/master/vulnerability/truncate-event-snapshots.md

    Note: this follows the specification very closely, but not exactly.
    """
    truncated = ""
    truncated_ranges = []
    curr_index = 0
    for tag in merged_tags:
        if curr_index < tag.start_index:
            truncated += truncate(str_representation[curr_index : tag.start_index])
        new_fragment = truncate(
            str_representation[tag.start_index : tag.end_index], length_factor=2
        )
        truncated += new_fragment
        truncated_ranges.append(
            Tag.from_range(len(truncated) - len(new_fragment), len(truncated))
        )
        curr_index = tag.end_index
    if curr_index < len(str_representation):
        truncated += truncate(str_representation[curr_index:])

    return truncated, truncated_ranges



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.agent.policy.patch_manager import reverse_patches_by_owner
from contrast.utils.patch_utils import (
    register_module_patcher,
    unregister_module_patcher,
)

MOD_WSGI_SERVER = "mod_wsgi.server"


def mod_wsgi_hook(mod_wsgi_server):
    from contrast.agent.agent_state import automatic_middleware
    from contrast.wsgi import ContrastMiddleware
    from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark

    def build___init__(orig_func, _):
        def __init__(wrapped, instance, args, kwargs):
            wrapped(*args, **kwargs)
            with automatic_middleware():
                instance.application = ContrastMiddleware(instance.application)

        return wrap_and_watermark(orig_func, __init__)

    build_and_apply_patch(
        mod_wsgi_server.ApplicationHandler,
        "__init__",
        build___init__,
    )


def register_patches():
    register_module_patcher(mod_wsgi_hook, MOD_WSGI_SERVER)


def reverse_patches():  # pragma: no cover
    unregister_module_patcher(MOD_WSGI_SERVER)
    module = sys.modules.get(MOD_WSGI_SERVER)
    if module is not None:
        reverse_patches_by_owner(module.ApplicationHandler)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from functools import cached_property

from contrast.reporting.teamserver_responses.protect_rule import ProtectRule


class ApplicationSettings:
    """
    This class is responsible for safely parsing V1 TeamServer Application Settings from a response to a usable format.
    The format can be found here: https://github.com/Contrast-Security-Inc/contrast-agent-api-spec. At the time of the
    creation of this class, the specific schema is ApplicationSettings in agent-endpoints.yml.
    """

    def __init__(self, application_settings_json: dict | None = None):
        self.application_settings_json = application_settings_json or {}

    @cached_property
    def disabled_assess_rules(self) -> list[str]:
        return [
            rule_name
            for rule_name, rule_details in self.application_settings_json.get(
                "assess", {}
            ).items()
            if not rule_details.get("enable", False)
        ]

    @cached_property
    def protect_rules(self) -> list[ProtectRule]:
        return [
            ProtectRule({"id": name, **value})
            for name, value in self.application_settings_json.get("protect", {})
            .get("rules", {})
            .items()
            if value
        ]

    @cached_property
    def sensitive_data_masking_policy(self) -> dict:
        return self.application_settings_json.get("sensitive_data_masking_policy", {})

    @cached_property
    def session_id(self) -> str | None:
        return self.application_settings_json.get("session_id", None)

    def common_config(self):
        ts_exclusions = self.application_settings_json.get("exclusions", {})
        return {
            "application.sensitive_data_masking_policy.mask_attack_vector": self.sensitive_data_masking_policy.get(
                "mask_attack_vector"
            ),
            "application.sensitive_data_masking_policy.mask_http_body": self.sensitive_data_masking_policy.get(
                "mask_http_body"
            ),
            "application.sensitive_data_masking_policy.rules": self.sensitive_data_masking_policy.get(
                "rules", []
            ),
            "application.url_exclusions": ts_exclusions.get("url", []),
            "application.input_exclusions": [
                {
                    # there's a hiccup in the API where "name" was repurposed for input exclusions
                    "modes": exclusion.get("modes", []),
                    "urls": exclusion.get("urls", []),
                    "match_strategy": exclusion.get("match_strategy", ""),
                    "assess_rules": exclusion.get("assess_rules", []),
                    "protect_rules": exclusion.get("protect_rules", []),
                    "input_type": exclusion.get("type", ""),
                    # there's a hiccup in the API where "name" was repurposed for input exclusions
                    "input_name": exclusion.get("name", ""),
                }
                for exclusion in ts_exclusions.get("input", [])
            ],
        }



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations


from collections.abc import Mapping
from os import environ
import sys
import atexit
import threading
import warnings
import sysconfig

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Literal, Optional
from contrast import policy_v2
from contrast.agent.framework import UNKNOWN_FRAMEWORK, Framework
from contrast.agent.policy import registry_v2
from contrast.utils.configuration_utils import get_platform

import contrast
from contrast import __version__, AGENT_CURR_WORKING_DIR, loader
from contrast.agent.exclusions import Exclusions
from contrast.agent.inventory import runtime_environment
from contrast.reporting import Reporter, get_reporting_client
from contrast.utils.locale import DEFAULT_ENCODING
from contrast.agent import (
    agent_lib,
    thread_watcher,
    patch_controller,
    sys_monitoring,
)
from contrast.agent.assess.sampling import SamplerConfig, update_sampler_config
from contrast.agent.assess import contrast_event
from contrast.agent.assess.rules.providers.enable import run_providers_in_thread
from contrast.agent.inventory.library_reader import LibraryReader
from contrast.agent.settings import Settings
from contrast.configuration.agent_config import AgentConfig
from contrast.utils.exceptions.deprecation_warning import ContrastDeprecationWarning
from contrast.utils.loggers.logger import (
    configure_syslog_logger,
    setup_basic_agent_logger,
    configure_agent_logger,
    setup_security_logger,
)
from contrast.utils import timer
from contrast.utils.namespace import Namespace
from contrast.utils.decorators import fail_loudly, fail_quietly
from contrast_rewriter import initialize_rewriter_logger
from contrast_rewriter import is_rewriter_enabled
from contrast.reporting.request_masker import RequestMasker

if TYPE_CHECKING:
    from contrast.agent.policy.registry_v2 import EventHandler


# NOTE: policy is currently loaded/generated on import. It is applied explicitly in
# policy/applicator.py
from contrast import policy  # noqa: F401

logger = setup_basic_agent_logger()
LOGS_SEPARATOR = "-" * 120

SHARED_WARNING_MESSAGE = """
For specific circumstances where explicit middleware configuration is still
required, set `agent.python.enable_automatic_middleware` to `false` in the
configuration file or set `CONTRAST__AGENT__PYTHON__ENABLE_AUTOMATIC_MIDDLEWARE=false`
in your environment to suppress this warning.

Please see the documentation for additional details:
    https://docs.contrastsecurity.com/en/python-middleware.html
"""

MIDDLEWARE_WITHOUT_SITECUSTOMIZE_WARNING = f"""
Explicit middleware configuration is no longer required or recommended. Instead,
users should run their application with the `contrast-python-run` command, which
automatically detects and enables framework-specific middleware.
{SHARED_WARNING_MESSAGE}
"""

NO_EXPLICIT_MIDDLEWARE_WARNING = f"""
Explicit middleware configuration is generally no longer necessary when using
the `contrast-python-run` command.
{SHARED_WARNING_MESSAGE}
"""

NO_SPECIFIC_MIDDLEWARE_WARNING = """
Using framework-specific middleware is no longer recommended for {}.
Use contrast.{}.ContrastMiddleware instead.
"""

PROTECT_MISMATCH_MESSAGE = (
    "Protect enabled in local config but disabled in Contrast UI. It is running."
)
ASSESS_MISMATCH_MESSAGE = (
    "Assess enabled in local config but disabled in Contrast UI. It is running."
)

if not contrast.telemetry_disabled():
    from contrast.agent.telemetry import Telemetry
else:
    Telemetry = None

Watcher = Optional  # A Watcher is a type that will watch config changes.


class module(Namespace):
    server_environment: str | None = None
    init_lock: threading.Lock = threading.Lock()
    sampling_cfg: Watcher[SamplerConfig] = None
    exclusions: Watcher[Exclusions] = None
    request_data_masker: Watcher[RequestMasker] = None
    event_handlers: dict[str, list[EventHandler]] = {}

    is_initialized = False
    id: int | None = None
    settings: Settings | None = None
    reporting_client: Reporter | None = None
    first_request: bool = True
    assess_enabled: bool = False
    assess_tags = ""
    observe_enabled: bool = False
    protect_enabled: bool = False

    automatic_middleware: ContextVar = ContextVar("automatic_middleware", default=False)

    deprecated_middleware: tuple[str, bool] | None = None

    configured_app_name: str | None = None
    configured_server_name: str | None = None

    # NOTE: these fields can be set prior to initialization
    framework: Framework = UNKNOWN_FRAMEWORK
    app_name: str | None = None


def is_runner_in_use() -> bool:
    return environ.get("CONTRAST_INSTALLATION_TOOL") == "CONTRAST_PYTHON_RUN"


def free_threading_available() -> bool:
    """
    Indicates if free-threading -could- be enabled, not whether or not it is actually
    enabled. The config option Py_GIL_DISABLED is slightly misleading.
    """
    if sys.version_info[:2] < (3, 13):
        return False
    return bool(sysconfig.get_config_var("Py_GIL_DISABLED"))


def free_threading_enabled() -> bool:
    """
    Indicates whether or not free-threading is actually enabled.
    """
    if sys.version_info[:2] < (3, 13):
        return False
    return not sys._is_gil_enabled()


@fail_quietly(return_value=False)
def jit_available() -> bool:
    """
    Indicates whether or not the JIT is built into this version of python.

    Potentially unreliable in 3.13.
    """
    if sys.version_info[:2] < (3, 13):
        return False
    if sys.version_info[:2] == (3, 13):
        return "_Py_JIT" in sysconfig.get_config_var("PY_CORE_CFLAGS")

    from sys import _jit

    return _jit.is_available()


@fail_quietly(return_value=False)
def jit_enabled() -> bool:
    """
    Indicates whether or not the JIT is currently enabled in this python process.

    Potentially unreliable in 3.13.
    """
    if sys.version_info[:2] < (3, 13):
        return False
    if sys.version_info[:2] == (3, 13):
        import _testinternalcapi

        get_optimizer = getattr(_testinternalcapi, "get_optimizer", lambda: None)
        return get_optimizer() is not None

    from sys import _jit

    return _jit.is_enabled()


@fail_quietly(return_value="unknown")
def tail_call_interpreter() -> bool | Literal["unknown"]:
    """
    Indicates whether or not the tail call interpreter is enabled in this python process.

    See https://docs.python.org/3.14/whatsnew/3.14.html#a-new-type-of-interpreter for details
    on the tail call interpreter.
    """
    return bool(sysconfig.get_config_var("Py_TAIL_CALL_INTERP"))


def _log_environment(settings: Settings):
    """
    Log current working directory, python version and pip version
    """
    log_level = settings.config.get_option("agent.logger.level").value().upper()

    logger.info(
        "ENVIRONMENT",
        python_version=sys.version,
        agent_version=__version__,
        assess_enabled=module.assess_enabled,
        protect_enabled=module.protect_enabled,
        using_runner=is_runner_in_use(),
        using_rewriter=is_rewriter_enabled(),
        preinstrument_flag=patch_controller.is_preinstrument_flag_set(),
        using_sys_monitoring=sys_monitoring.is_enabled(),
        log_level=log_level,
        configured_application_name=module.configured_app_name,
        detected_application_name=module.app_name,
        detected_framework=module.framework,
        discovered_webserver=settings.server,
        cwd=AGENT_CURR_WORKING_DIR,
        executable=sys.executable,
        platform=get_platform(),
        free_threading_available=free_threading_available(),
        free_threading_enabled=free_threading_enabled(),
        jit_available=jit_available(),
        jit_enabled=jit_enabled(),
        tail_call_interpreter=tail_call_interpreter(),
        default_encoding=DEFAULT_ENCODING,
        debug_info={"sys.argv": sys.argv} if log_level == "DEBUG" else None,
    )

    # py39: remove
    if sys.version_info[:2] == (3, 9):
        _log_and_warn(
            "Contrast Security support for Python 3.9 is deprecated "
            "and will be removed in October 2025"
        )


def _warn_for_misleading_config(settings: Settings):
    protect_option = settings.config.get_option("protect.enable")
    protect_enabled = protect_option.value()
    if protect_enabled and not protect_option.ui_value:
        logger.warning(PROTECT_MISMATCH_MESSAGE)

    assess_option = settings.config.get_option("assess.enable")
    assess_enabled = assess_option.value()
    if assess_enabled and not assess_option.ui_value:
        logger.warning(ASSESS_MISMATCH_MESSAGE)

    if (
        settings.is_agent_config_enabled()
        and not protect_enabled
        and not assess_enabled
    ):
        logger.warning("Neither Protect nor Assess is enabled. Neither is running.")


def add_config_watchers(config: AgentConfig) -> None:
    """
    Adds all config watchers to the config object.
    """
    config.add_watcher(configure_agent_logger)
    config.add_watcher(configure_syslog_logger)
    config.add_watcher(set_assess_enabled)
    config.add_watcher(set_observe_enabled)
    config.add_watcher(set_protect_enabled)
    config.add_watcher(set_exclusions)
    config.add_watcher(set_data_masking)
    config.add_watcher(set_server_environment_and_sampler)


def initialize_libraries(reporter: Reporter):
    """
    Read libraries from the application
    """
    LibraryReader(reporter).start()


@fail_loudly("Unable to initialize Contrast Agent Settings.", return_value=False)
def initialize_settings():
    """
    Initialize agent settings.

    Returns True on settings being initialized and False if any failure.
    """
    module.settings = Settings()

    return True


def is_initialized() -> bool:
    return module.is_initialized


def get_settings():
    return module.settings


def get_server_environment():
    return module.server_environment


def _get_app_name(config: Mapping, app_name):
    if config and (config_name := config.get("application.name")):
        return config_name

    return app_name if app_name else "root"


def get_app_name():
    return module.configured_app_name or _get_app_name(
        module.settings.config, module.app_name
    )


def _get_server_name(config: Mapping):
    """
    Hostname of the server

    Default is socket.gethostname() or localhost
    """
    if config:
        return config.get("server.name")
    return None


def get_server_name():
    return module.configured_server_name or _get_server_name(module.settings.config)


def is_first_request():
    return module.first_request


def set_first_request(val: bool):
    module.first_request = val


def set_assess_enabled(config: Mapping):
    module.assess_enabled = config["assess.enable"]


def set_protect_enabled(config: Mapping):
    module.protect_enabled = config["protect.enable"]


def set_observe_enabled(config: Mapping):
    module.observe_enabled = config["observe.enable"]


@fail_loudly("Failed to register detected framework")
def set_detected_framework(distribution_name: str):
    """
    Registers the detected framework by its distribution name.

    It is safe to call this method prior to initialization of agent state. In
    order to be effective, this method should be called prior to middleware
    initialization.
    """
    module.framework = Framework(distribution_name)


def detected_framework() -> Framework:
    return module.framework


@fail_loudly("Failed to register application name")
def set_application_name(name: str | None):
    """
    Registers the name of the application

    It is safe to call this method prior to initialization of agent state. In
    order to be effective, this method should be called prior to middleware
    initialization.
    """
    module.app_name = name


def set_server_environment_and_sampler(config):
    module.server_environment = config.get("server.environment")

    update_sampler_config(config)


def set_exclusions(config: Mapping):
    module.exclusions = Exclusions(config)


def set_data_masking(config: Mapping):
    module.request_data_masker = RequestMasker(config)


def _log_and_warn(msg):
    warnings.warn(msg, ContrastDeprecationWarning, stacklevel=2)
    logger.warning(msg)


def _check_middleware_warnings():
    # NOTE: currently, we are skipping framework-specific middleware deprecation
    # warnings. In the future, we might want to actually deprecate these. For now,
    # however, we are holding off.
    # this should be `if module.deprecated_middleware is not None:`
    if False:
        (  # pylint: disable=unpacking-non-sequence
            framework,
            is_asgi,
        ) = module.deprecated_middleware
        message = NO_SPECIFIC_MIDDLEWARE_WARNING.format(
            framework, "asgi" if is_asgi else "wsgi"
        )
        _log_and_warn(message)

    # This configuration suppresses both warnings
    if not module.settings.config.enable_automatic_middleware:
        return

    # Warn for the case where middleware is only explicitly configured,
    # missing sitecustomize early rewrites and patches.
    if not loader.SITECUSTOMIZE_LOADED:
        _log_and_warn(MIDDLEWARE_WITHOUT_SITECUSTOMIZE_WARNING)
    # Warn for the case where middleware is explicitly configured but also detected automatically
    # An actual warning doesn't seem necessary in this case, just a log message
    elif not module.automatic_middleware.get():
        logger.warning(NO_EXPLICIT_MIDDLEWARE_WARNING)


def initialize():
    """
    If this method is called more than once per process, we use the is_initialized
    flag to only run the following work once:
        - library analysis thread initialization
        - turning on patches
        - hardcoded rule providers
        - scanning config rules
        - environment logging
        - loading common configuration
        - initializing settings

    In order to avoid a warning message, callers should check is_initialized()
    prior to calling this function.
    """
    with module.init_lock:
        if module.is_initialized:
            _check_middleware_warnings()
            logger.warning("Attempted to initialize agent state more than once")
            return

        module.id = id(module)

        logger.info(
            "Initializing Contrast Agent %s",
            __name__,
            id=module.id,
            version=__version__,
        )

        if not initialize_settings():
            return

        if module.settings is None:
            logger.warning("Contrast Agent is not enabled.")
            return

        module.configured_server_name = _get_server_name(module.settings.config)
        module.configured_app_name = _get_app_name(
            module.settings.config, module.app_name
        )
        module.assess_tags = module.settings.config["assess.tags"]

        if not module.settings.is_agent_config_enabled():
            logger.warning("Contrast Agent is disabled by local configuration.")
            module.settings.log_effective_config()
            return

        add_config_watchers(module.settings.config)
        setup_security_logger(module.settings.config)
        module.settings.config.log_config()

        # NOTE (TODO: PYT-3016)
        # This assumes that initialize is only ever called from middleware.
        # If at some point this no longer holds, it may be necessary to pass a
        # flag to initialize indicating whether these warnings should be
        # processed or not.
        _check_middleware_warnings()

        # The rewriter is applied long before we have any settings or logger so
        # the deferred logs are finally processed here
        initialize_rewriter_logger(logger)

        module.reporting_client = get_reporting_client(module.settings.config)
        module.reporting_client.start()

        if not module.reporting_client.initialize_application(
            module.settings.config,
            server_type=module.settings.server_type,
        ):
            disable_agent(reason="Unable to initialize application in Contrast UI")
            module.settings.log_effective_config()
            return

        _warn_for_misleading_config(module.settings)

        preinstrument = patch_controller.is_preinstrument_flag_set()

        if module.assess_enabled or preinstrument:
            sys_monitoring.enable()

        # Eventually this should be handled by Fireball.
        from contrast.reporting.teamserver_messages.effective_config import (
            EffectiveConfig,
        )

        module.reporting_client.add_message(EffectiveConfig())

        # This MUST happen after the initialization calls for TeamServer messaging
        # (sending server start and application start) to ensure that TeamServer will
        # accept the messages sent by our background reporting threads
        # Skip telemetry since it is enabled later in this method
        thread_watcher.ensure_running(module, skip_telemetry=True)

        module.settings.log_effective_config()

        if (
            module.settings.is_inventory_enabled() or preinstrument
        ) and module.settings.is_analyze_libs_enabled():
            initialize_libraries(module.reporting_client)

        runtime_environment.report_server_runtime(
            module.reporting_client,
            module.settings.config.get("server.discover_cloud_resource"),
        )

        registry_v2.register_policy_definitions(policy_v2.definitions())
        module.event_handlers = registry_v2.generate_policy_event_handlers(
            assess=module.assess_enabled,
            observe=module.observe_enabled,
            protect=module.protect_enabled,
        )

        patch_controller.enable_patches(preinstrument=preinstrument)

        if module.assess_enabled or preinstrument:
            # For now agent runtime starts before config scanning
            # this will be reset when time_limit_threshold is reached,
            # it doesn't symbolize the total agent runtime for all time.
            module.settings.agent_runtime_window = timer.now_ms()

            contrast_event.initialize(module.settings.config)

            run_providers_in_thread()
            # NOTE (TODO: PYT-3016) This should stay in middleware
            # scan_configs_in_thread()

        _log_environment(module.settings)

        if (module.protect_enabled or preinstrument) and not agent_lib.initialize(
            module.settings.config
        ):
            disable_agent(reason="Unable to initialize Protect analysis library")
            return

        if free_threading_enabled():
            disable_agent(
                reason=(
                    "Free-threading is enabled in this python process - Contrast cannot"
                    " run in this environment"
                )
            )
            return

        if jit_enabled():
            disable_agent(
                reason=(
                    "Python's experimental JIT compiler is enabled in this process -"
                    " Contrast cannot run in this environment"
                )
            )
            return

        logger.info(
            "Finished Initializing Contrast Agent %s",
            __name__,
            id=module.id,
        )

        if Telemetry is not None:
            contrast.TELEMETRY = Telemetry()
            contrast.TELEMETRY.start()
            atexit.register(contrast.TELEMETRY.stop)

        module.is_initialized = True


def disable_agent(reason: str):
    logger.error("Fatal error: Disabling Contrast", reason=reason)
    if module.settings and module.settings.config:
        enable_option = module.settings.config.get_option("enable")
        enable_option.override_value = False


def detected_deprecated_middleware(*, framework: str = "", is_asgi: bool = False):
    module.deprecated_middleware = (framework, is_asgi)


@contextmanager
def automatic_middleware():
    """
    Context manager to be used by automatic middleware hooks when initializing middleware

    This enables us to determine when middleware is initialized from within our
    automatic middleware hooks without having to change the API of any
    middleware classes.
    """
    try:
        module.automatic_middleware.set(True)
        yield
    finally:
        module.automatic_middleware.set(False)


def in_automatic_middleware() -> bool:
    return module.automatic_middleware.get()



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators import BasePropagator
from contrast.agent.assess.utils import (
    copy_events,
    copy_tags_in_span,
    get_properties,
)


class SplitPropagator(BasePropagator):
    PARTITION_METHODS = ["partition", "rpartition"]
    REVERSE_METHODS = ["rsplit", "rpartition"]

    def track_target(self):
        # NOP. Let propagate handle the decision to track
        pass

    @property
    def needs_propagation(self):
        return self.any_source_tracked

    def build_event(self, target_properties, tagged):
        parents = [self.source_properties.event] if self.source_properties.event else []
        # For split, the tagged individual string is passed when building the
        # event, but the target is actually the array of strings that was
        # returned by split.
        target_properties.build_event(
            self.node,
            tagged,
            self.preshift.obj,
            self.target,
            self.preshift.args,
            self.preshift.kwargs,
            parents,
            None,
        )

    def add_tags_and_properties(self, ret):
        if self.source_properties is None:
            return

        for target in self.target:
            if self.node.tags:
                self.apply_tags(self.node, target)

            if self.node.untags:
                self.apply_untags(self.node, target)

            target_properties = get_properties(target)
            if target_properties is None:
                continue

            target_properties.add_properties(self.node.properties)

            self.build_event(target_properties, target)

    def propagate(self):
        source = self.sources[0]
        self.source_properties = get_properties(source)
        if self.source_properties is None:
            return

        # Offset in the original string
        source_offset = 0

        reverse = self.node.method_name in self.REVERSE_METHODS

        # The target (result of split) is an array of strings
        target = self.target[::-1] if reverse else self.target

        if self.node.method_name in self.PARTITION_METHODS:
            partition = self.preshift.args[0]
        else:
            partition = None

        for newstr in target:
            if reverse:
                source_offset = source.rfind(newstr)
                source = source[:source_offset]
            else:
                source_offset = source.find(newstr, source_offset)

            if partition and newstr == partition:
                # Make sure that we don't skip another string that just happens
                # to be the same as the partition string.
                partition = None
                continue

            span = (source_offset, source_offset + len(newstr))
            copy_tags_in_span(newstr, self.source_properties, span)

            target_properties = get_properties(newstr)
            if target_properties is not None:
                copy_events(target_properties, self.source_properties)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes


serialize_propagators = [
    {
        "module": ["json", "simplejson"],
        "method_name": "dumps",
        "source": "ARG_0,KWARG:obj",
        "target": "RETURN",
        "action": "JSON",
    },
    {
        "module": ["json", "simplejson"],
        "method_name": "loads",
        "source": "ARG_0,KWARG:s",
        "target": "RETURN",
        "action": "JSON",
    },
    {
        "module": "pickle",
        "method_name": "dumps",
        "source": "ARG_0,KWARG:obj",
        "target": "RETURN",
        "action": "SPLAT",
    },
    {
        "module": "yaml",
        "method_name": "dump",
        "source": "ARG_0,KWARG:data",
        "target": "RETURN",
        "action": "SPLAT",
    },
    {
        "module": "yaml",
        "method_name": "dump_all",
        "source": "ARG_0,KWARG:documents",
        "target": "RETURN",
        "action": "SPLAT",
    },
]


register_propagation_nodes(serialize_propagators)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.


class HttpOnlyRuleMixin:
    @property
    def name(self):
        return "httponly"

    def is_violated(self, value):
        """
        The rule is violated if the value is False or if it is not set at all (None)
        """
        if self.count_threshold_reached():
            return False
        return not value



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from . import BasePropagator


class Tagger(BasePropagator):
    """
    Propagation action that only applies tags/untags

    This is different from KEEP, which also copies tags from the source(s) to the target.
    """

    def propagate(self):
        return



import base64
from collections.abc import Generator
import json
import os
from pathlib import Path
import platform
import shutil

import pytest
from contrast.version import __version__ as AGENT_VERSION
from syrupy.assertion import SnapshotAssertion
from syrupy.extensions.json import JSONSnapshotExtension
from testcontainers.core.image import DockerImage
from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import re, wait_for_logs
from testcontainers.generic import ServerContainer


API_TOKEN = os.environ.get("CONTRAST__API__TOKEN", "")
if not API_TOKEN:
    legacy_keys = {
        key_name: os.environ.get(f"CONTRAST__API__{key_name.upper()}")
        for key_name in [
            "api_key",
            "service_key",
            "url",
            "user_name",
        ]
    }
    missing_keys = [key_name for key_name, value in legacy_keys.items() if not value]
    if missing_keys:
        pytest.skip(
            reason="API token is not set and required legacy keys are missing: "
            f"{', '.join(missing_keys)}"
        )
    else:
        API_TOKEN = base64.b64encode(bytes(json.dumps(legacy_keys), "utf-8")).decode()

OBSERVE_APP_NAME = (
    f"observe-django-integration-test-app-{AGENT_VERSION}-{platform.python_version()}"
)
if summary_file_name := os.environ.get("GITHUB_STEP_SUMMARY"):
    with open(summary_file_name, "a") as summary_file:
        summary_file.write(
            f"## Observe Django Integration Test App Name\n\n{OBSERVE_APP_NAME}\n\n"
        )


@pytest.fixture
def snapshot_json(snapshot: SnapshotAssertion) -> SnapshotAssertion:
    return snapshot.use_extension(JSONSnapshotExtension)


@pytest.fixture(scope="module")
def mitmproxy(
    shared_network: Network, clean_log_dir: Path
) -> Generator[ServerContainer, None, None]:
    with DockerImage(local_path("mitmproxy")) as image:
        REQUEST_AUDIT_DIR = "/home/mitmproxy/request_audit_logs"
        proxy_server = (
            ServerContainer(port=6066, image=image)
            .with_env("REQUEST_AUDIT_DIR", REQUEST_AUDIT_DIR)
            .with_volume_mapping(clean_log_dir, REQUEST_AUDIT_DIR, mode="rw")
            .with_volume_mapping("mitmproxy-certs", "/root/.mitmproxy", mode="rw")
            .with_network(shared_network)
            .with_network_aliases("mitmproxy")
        )
        # HACK: ServerContainer is going to try to connect to "/" when
        # it enters the context manager. This is going to fail because
        # mitmproxy forwards this request to localhost at the exposed port,
        # which doesn't have a web server running.
        #
        # We override the _connect method to avoid this.
        proxy_server._connect = lambda: ...
        with proxy_server:
            wait_for_logs(proxy_server, re.escape("proxy listening at"))

            yield proxy_server


@pytest.fixture(scope="module")
def observe_django_app(
    mitmproxy: ServerContainer, shared_network: Network, clean_log_dir: Path
) -> Generator[ServerContainer, None, None]:
    with (
        DockerImage(local_path("apps", "django"), target="instrumented-app") as image,
        ServerContainer(port=8080, image=image)
        .with_envs(
            CONTRAST_PUBLIC_BUILD="false",
            CONTRAST_TESTING="true",
            CONTRAST__API__PROXY__ENABLE="true",
            CONTRAST__API__PROXY__URL="http://mitmproxy:6066",
            CONTRAST__API__CERTIFICATE__CA_FILE="/etc/mitmproxy/mitmproxy-ca.pem",
            CONTRAST__API__TOKEN=API_TOKEN,
            CONTRAST__OBSERVE__ENABLE="true",
            CONTRAST__AGENT__POLLING__APP_ACTIVITY_MS="5000",
            CONTRAST__REPORTING__OBSERVE__PROTOCOL="http_json",
            CONTRAST__APPLICATION__NAME=OBSERVE_APP_NAME,
            CONTRAST__SERVER__NAME="observe-integration-tests",
        )
        .with_volume_mapping("mitmproxy-certs", "/etc/mitmproxy")
        .with_volume_mapping(local_path(".."), "/agent", mode="rw")
        .with_volume_mapping(clean_log_dir, "/tmp/logs", mode="rw")
        .with_network(shared_network) as app_server,
    ):
        wait_for_logs(app_server, re.escape("Quit the server with CONTROL-C."))
        app_server.get_api_url = app_server._create_connection_url

        yield app_server


def local_path(*args: str) -> Path:
    """Returns an absolute path relative to this directory."""
    return Path(__file__, "..", *args).resolve()


@pytest.fixture(scope="module")
def shared_network() -> Generator[Network, None, None]:
    with Network() as network:
        yield network


@pytest.fixture(scope="module")
def clean_log_dir() -> Path:
    """
    Fixture to ensure the log directory is clean before tests run.
    """
    log_dir = local_path("logs")
    shutil.rmtree(local_path("logs"), ignore_errors=True)
    return log_dir



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
# pylint: disable=too-many-lines
from __future__ import annotations
from collections import defaultdict
from datetime import datetime, timezone
from email.utils import format_datetime
from functools import cached_property
import json
import os
import pathlib
from urllib.parse import urlparse

from contrast import AGENT_CURR_WORKING_DIR

from contrast.agent import scope
from contrast.utils.decorators import fail_quietly
from contrast.agent.framework import _ServerTypeFramework, Server
from contrast.agent.reaction_processor import ReactionProcessor
from contrast.configuration.config_option import (
    CONTRAST_UI_SRC,
    USER_CONFIGURATION_FILE_SRC,
    DEFAULT_VALUE_SRC,
)
from contrast.configuration.agent_config import AgentConfig
from contrast.utils.loggers.logger import (
    STDOUT,
    STDERR,
)
from contrast.utils.singleton import Singleton
from contrast.utils.string_utils import truncate
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

ASSESS_STACKTRACES = "assess.stacktraces"
ASSESS_DISABLED_RULE_CONFIG_KEY = "assess.rules.disabled_rules"
EXPORTED_CONFIG_FILE_NAME = "contrast_connection.json"

UNIX_EPOCH_HTTP = format_datetime(
    datetime.fromtimestamp(0, tz=timezone.utc), usegmt=True
)
"""
String representation of the unix epoch (1 January 1970) in HTTP-format. Used in the
`If-Modified-Since` header before we have a `Last-Modified` time.
"""


def apply_server_settings(
    config: AgentConfig,
    server_settings_json: dict,
    settings: Settings | None = None,
):
    """
    Applies v1 server settings from the UI to config and settings.

    Eventually, all changes should go through config so that settings
    will be removed.
    """
    from contrast.reporting.teamserver_responses.server_settings import (
        ServerSettings,
    )

    server_features = ServerSettings(server_settings_json)
    ui_config = server_features.common_config()
    logger.debug("Received updated server features", config=ui_config)
    if _changed := config.update_ui_values(ui_config):
        config.log_config()

    if settings:
        settings.set_protect_rules()
        settings.last_server_update_time_ms = now_ms()


def apply_ts_feature_settings(
    config: AgentConfig,
    response_body: dict,
    settings: Settings | None = None,
):
    """
    Applies NG server settings from the UI to config and settings.

    Eventually, all changes should go through config so that settings
    will be removed.
    """
    from contrast.reporting.teamserver_responses.ng_server_settings import (
        NGServerSettings,
    )

    server_features = NGServerSettings(response_body)
    ui_config = server_features.common_config()
    logger.debug("Received updated server features", config=ui_config)
    if _changed := config.update_ui_values(ui_config):
        config.log_config()

    if settings:
        settings.set_protect_rules()
        settings.last_server_update_time_ms = now_ms()


def _apply_application_settings(
    config: AgentConfig,
    application_settings_json: dict,
    settings: Settings | None,
    last_modified: str | None,
):
    """
    Applies v1 app settings from the UI to config and settings.

    Eventually, all changes should go through config so that settings
    will be removed.
    """
    from contrast.reporting.teamserver_responses.application_settings import (
        ApplicationSettings,
    )

    app_settings = ApplicationSettings(application_settings_json)
    ui_config = app_settings.common_config()
    logger.debug("Received updated application settings", config=ui_config)

    if _changed := config.update_ui_values(ui_config):
        config.log_config()

    if settings:
        settings.config.set_ui_value(
            ASSESS_DISABLED_RULE_CONFIG_KEY, app_settings.disabled_assess_rules
        )
        # This is the only place session_id is set by TS.
        # If session id is set in the config, that value will be echoed back by TS
        if app_settings.session_id:
            settings.config.set_ui_value(
                "application.session_id", app_settings.session_id
            )
        settings.set_protect_rules()
        known_rules = set(settings.protect_rules)
        logger.debug("Checking for rules: %s", known_rules)
        if app_settings.protect_rules:
            for definition in app_settings.protect_rules:
                logger.debug("Definition: %s  --- %s", definition.id, definition.mode)
                settings.config.set_ui_value(
                    f"protect.rules.{definition.id}.mode", definition.mode
                )
                if definition.id in known_rules:
                    known_rules.remove(definition.id)
        logger.debug("Remaining rules: %s", known_rules)
        for rule_name in known_rules:
            settings.config.set_ui_value(f"protect.rules.{rule_name}.mode", None)
        settings.last_app_update_time_ms = now_ms()
        if last_modified:
            settings.app_settings_last_modified = last_modified


class Settings(Singleton):
    @scope.contrast_scope()
    def init(self, framework_name: str | None = None):
        """
        Agent settings for the entire lifetime of the agent.

        Singletons should override init, not __init__.
        """
        # We need to initialize config to None because other methods
        # on Settings are called within this init function. It's
        # not obvious because these calls come through import patches.
        self.config = None

        self.config_features = {}
        self.last_server_update_time_ms = 0
        self.last_app_update_time_ms = 0
        # HTTP-style timestamp used in `If-Modified-Since` / `Last-Modified` headers
        # Generally, the AgentStartup response will give us a real value here before we
        # need to send this in a request to teamserver, but we have the unix epoch as a
        # fallback just in case.
        self.app_settings_last_modified: str = UNIX_EPOCH_HTTP
        self.heartbeat = None
        self.server_settings_thread = None
        self.application_settings_thread = None
        self.server = Server()
        self.sys_module_count = 0

        # Server
        self.server_path = None
        self.server_type = None

        # Rules
        self.protect_rules = dict()

        # circular import
        from contrast.agent.assess.rules.response.autocomplete_missing_rule import (
            AutocompleteMissingRule,
        )
        from contrast.agent.assess.rules.response.cache_controls_rule import (
            CacheControlsRule,
        )
        from contrast.agent.assess.rules.response.clickjacking_rule import (
            ClickjackingRule,
        )

        from contrast.agent.assess.rules.response.x_content_type_rule import (
            XContentTypeRule,
        )

        from contrast.agent.assess.rules.response.csp_header_missing_rule import (
            CspHeaderMissingRule,
        )

        from contrast.agent.assess.rules.response.x_xss_protection_disabled_rule import (
            XXssProtectionDisabledRule,
        )

        from contrast.agent.assess.rules.response.hsts_header_rule import HstsHeaderRule

        from contrast.agent.assess.rules.response.csp_header_insecure_rule import (
            CspHeaderInsecureRule,
        )

        from contrast.agent.assess.rules.response.parameter_pollution_rule import (
            ParameterPollutionRule,
        )

        self.assess_response_rules = [
            AutocompleteMissingRule(),
            CacheControlsRule(),
            ClickjackingRule(),
            XContentTypeRule(),
            CspHeaderMissingRule(),
            XXssProtectionDisabledRule(),
            HstsHeaderRule(),
            CspHeaderInsecureRule(),
            ParameterPollutionRule(),
        ]

        # Initialize config
        self.config = AgentConfig()

        self.server_type = (
            self.config["server.type"]
            or _ServerTypeFramework(framework_name).name_lower
        )

        self.agent_runtime_window = now_ms()

        logger.info("Contrast Agent finished loading settings.")

    @cached_property
    def is_proxy_enabled(self):
        return self.config.get("api.proxy.enable")

    @cached_property
    def proxy_url(self):
        return self.config.get("api.proxy.url")

    @cached_property
    def api_url_scheme(self):
        scheme = urlparse(self.api_url).scheme
        return scheme

    @property
    def ignore_cert_errors(self):
        return self.config.get("api.certificate.ignore_cert_errors")

    @property
    def ca_file(self):
        return self.config.get("api.certificate.ca_file")

    @property
    def client_cert_file(self):
        return self.config.get("api.certificate.cert_file")

    @property
    def client_private_key(self):
        return self.config.get("api.certificate.key_file")

    @cached_property
    def api_service_key(self):
        return self.config.get("api.service_key")

    @cached_property
    def api_url(self):
        """Normalizes the URL to remove any whitespace or trailing slash"""
        return self.config.get("api.url").strip().rstrip("/")

    @cached_property
    def api_key(self):
        return self.config.get("api.api_key")

    @cached_property
    def api_user_name(self):
        return self.config.get("api.user_name")

    def is_agent_config_enabled(self):
        if self.config is None:
            return True

        return self.config.get("enable")

    @cached_property
    def max_sources(self):
        return self.config.get("assess.max_context_source_events")

    @cached_property
    def max_propagation(self):
        return self.config.get("assess.max_propagation_events")

    @cached_property
    def max_vulnerability_count(self):
        """Max number of vulnerabilities per rule type to report for one
        agent run `time_limit_threshold` time period"""
        return self.config.get("assess.max_rule_reported")

    @cached_property
    def agent_runtime_threshold(self):
        return self.config.get("assess.time_limit_threshold")

    @cached_property
    def app_path(self):
        return self.config.get("application.path")

    @cached_property
    def app_version(self):
        return self.config.get("application.version")

    @property
    def pid(self):
        """
        pid is used in our CMDi protect rule.

        pid must be unique for each worker process of an app.
        :return: int current process id
        """
        return os.getpid()

    def establish_heartbeat(self, reporting_client):
        """
        Initialize Heartbeat between Agent and TS if it has not been already initialized.
        """
        if self.heartbeat is None:
            # Circular import
            from contrast.agent.heartbeat_thread import HeartbeatThread

            self.heartbeat = HeartbeatThread(reporting_client)
            self.heartbeat.start()

    def establish_server_settings_thread(self, reporting_client):
        """
        Initialize Server Settings poll between Agent and TS if it has not been already
        initialized.
        """
        if self.server_settings_thread is None:
            # Circular import
            from contrast.agent.settings_threads import ServerSettingsThread

            self.server_settings_thread = ServerSettingsThread(reporting_client)
            self.server_settings_thread.start()

    def establish_application_settings_thread(self, reporting_client):
        """
        Initialize Application Settings poll between Agent and TS if it has not been
        already initialized.
        """
        if self.application_settings_thread is None:
            # Circular import
            from contrast.agent.settings_threads import ApplicationSettingsThread

            self.application_settings_thread = ApplicationSettingsThread(
                reporting_client
            )
            self.application_settings_thread.start()

    def apply_ts_feature_settings(self, response_body):
        apply_ts_feature_settings(self.config, response_body, self)

    def process_ts_reactions(self, response_body):
        # App startup/activity wrap reactions in a settings dict whereas
        # Server startup/activity has it at the top level response dict
        reactions = response_body.get("settings", {}).get("reactions", None)

        if not reactions:
            reactions = response_body.get("reactions", None)

        if not reactions:
            return

        ReactionProcessor.process(reactions, self)

    def apply_application_settings(
        self, application_settings_json: dict, last_modified: str | None = None
    ):
        """
        Update stored application settings using the provided dict. This corresponds to
        the teamserver v1 API ApplicationSettings1.0 object.

        `last_modified` is an HTTP-style timestamp from the `Last-Modified` header. If
        set to None, the previous known last update time will be maintained; note that
        the last update time is initially set to the unix epoch.
        """
        _apply_application_settings(
            self.config, application_settings_json, self, last_modified
        )

    def apply_identification(self, identification_json: dict | None):
        if not identification_json:
            return
        self.application_uuid = identification_json.get("application_uuid", None)
        self.organization_uuid = identification_json.get("organization_uuid", None)
        self.server_uuid = identification_json.get("server_uuid", None)
        self.session_id = identification_json.get("session_id", None)

    def apply_server_settings(self, server_settings_json: dict):
        apply_server_settings(self.config, server_settings_json, self)

    def is_inventory_enabled(self):
        """
        inventory.enable = false: Disables both route coverage and library analysis and reporting
        """
        return self.config.get("inventory.enable")

    def is_analyze_libs_enabled(self):
        """
        inventory.analyze_libraries = false: Disables only library analysis/reporting
        """
        return (
            self.config is not None
            and self.config.get("inventory.analyze_libraries")
            and self.is_inventory_enabled()
        )

    def set_protect_rules(self):
        from contrast.agent.protect.rule.rules_builder import build_protect_rules

        self.protect_rules = build_protect_rules()

    def get_server_path(self):
        """
        Working Directory of the server

        Default is root
        """
        if self.server_path is None:
            self.server_path = self.config.get("server.path") or truncate(
                AGENT_CURR_WORKING_DIR
            )

        return self.server_path

    def is_assess_rule_disabled(self, rule_id):
        """
        Rules disabled in config override all disabled rules from TS per common config
        """
        disabled_rules = self.config.get(ASSESS_DISABLED_RULE_CONFIG_KEY)
        return disabled_rules is not None and rule_id in disabled_rules

    def enabled_response_rules(self):
        return (
            [
                rule
                for rule in self.assess_response_rules
                if not self.is_assess_rule_disabled(rule.name)
            ]
            if self.config.get("assess.enable_scan_response")
            else []
        )

    def is_collect_stacktraces_all(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "ALL"

    def is_collect_stacktraces_some(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "SOME"

    def is_collect_stacktraces_none(self):
        return self.config is not None and self.config.get(ASSESS_STACKTRACES) == "NONE"

    def build_proxy_url(self):
        if self.proxy_url:
            # self.api_url_scheme is the key of the scheme we are proxying
            # self.proxy_url is the actual proxy url and port
            # https://requests.readthedocs.io/en/latest/user/advanced/#proxies
            return {self.api_url_scheme: self.proxy_url}

        return {}

    @cached_property
    def contrast_ui_status(self):
        config_name = "enable"
        error_msg = "Unable to connect to Contrast; configuration details from the Contrast UI will not be included."

        enable_option = self.config.get_option(config_name)
        if enable_option is None or enable_option.ui_value is False:
            # This indicates we were told to shut down by Contrast UI
            return error_msg

        return None

    def generate_effective_config(self):
        """
        The purpose of this function is to export as many of the configuration values found in AgentConfig
        to the effective_config list. This list will be logged to disk. Some values that where set using
        the Contrast UI where saved to AgentConfig in attempt to have a single place storing all settings.

        In some cases this wasn't immediately feasible due to when and how we reload those settings
        """
        status = self.contrast_ui_status or self.config.config_status or "Success"
        report_creation_time = datetime.now(timezone.utc).strftime(RFC3339_FORMAT)
        config = {
            "report_create": report_creation_time,
            "config": {
                "status": status,
                "effective_config": [],
                "user_configuration_file": [],
                "environment_variable": [],
                "contrast_ui": [],
            },
        }

        effective_config = config["config"]["effective_config"]
        env_config = config["config"]["environment_variable"]
        ui_config = config["config"]["contrast_ui"]

        user_config = defaultdict(list)

        for key in self.config:
            option = self.config.get_option(key)
            if option is None or not option.log_effective_config:
                continue

            # The spec states this value is not included if we can't communicate with Contrast UI
            if (
                option.source() == CONTRAST_UI_SRC
                and self.contrast_ui_status is not None
            ):
                continue

            value = option.loggable_value()

            # If the value is unset and has no default value, we should skip it.
            # For instance, if a proxy isn't set, there's no reason to report those configuration values.
            if option.source() == DEFAULT_VALUE_SRC and value in ("", "[]"):
                continue

            config_entry = {
                "canonical_name": option.canonical_name,
                "name": option.provided_name(),
                "source": option.source(),
                "value": value,
            }
            if option.source() == USER_CONFIGURATION_FILE_SRC:
                config_entry["filename"] = option.file_name()

            effective_config.append(config_entry)

            if option.env_value is not None:
                config_entry = {
                    "canonical_name": option.canonical_name,
                    "name": option.provided_name(),
                    "value": option.to_string(option.env_value),
                }
                env_config.append(config_entry)
            if option.ui_value is not None:
                config_entry = {
                    "canonical_name": option.canonical_name,
                    # provided name is only for env vars
                    "name": option.canonical_name,
                    "value": option.to_string(option.ui_value),
                }
                ui_config.append(config_entry)
            if option.file_values:
                for val, filename in zip(option.file_values, option.file_sources):
                    config_entry = {
                        "canonical_name": option.canonical_name,
                        # provided name is only for env vars
                        "name": option.canonical_name,
                        "value": option.to_string(val),
                    }
                    user_config[filename].append(config_entry)

        config["config"]["user_configuration_file"] = [
            {"path": filename, "values": values}
            for filename, values in user_config.items()
        ]

        # PYT-3808: Don't send properties that are null or empty lists
        config["config"] = {
            k: v for k, v in config["config"].items() if v not in ([], None)
        }

        return config

    @fail_quietly("Failed to export effective config")
    def log_effective_config(self):
        """
        If we have a directory to which to log safely, we are to add the effective configuration json file to that
        location. For other instances, such as STDOUT or un-writable directories, there is no need to write the file as
        the configuration is contained in our standard logs.
        """
        if (
            self.config.get("agent.logger.stdout")
            or self.config.get("agent.logger.stderr")
            or self.config.get("agent.logger.path") in (STDOUT, STDERR)
        ):
            return
        path = pathlib.Path(self.config.get("agent.logger.path")).parent.resolve()
        filename = f"{path}/{EXPORTED_CONFIG_FILE_NAME}"

        try:
            with open(filename, "w") as f:
                json.dump(self.generate_effective_config(), f, indent=4)
        except PermissionError as perm_error:
            logger.warning(
                "Failed to export effective config",
                filename=filename,
                error=str(perm_error),
            )
            return


RFC3339_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_SQL_INJECTION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_SQL_INJECTION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "SQL_ENCODED",
]

# NOTE: The nodes for the methods in dbapi2 modules do not correspond exactly to real
# patch locations. Instead, we pretend that every dbapi2 cursor object is
# called Cursor. This is done for convenience and for storytelling purposes.
# Our patches (which in these cases are not controlled by policy) account for
# this and perform the policy lookup correctly
sql_injection_triggers = [
    {
        "module": ["sqlite3.dbapi2", "sqlite3"],
        "class_name": "Cursor",
        "method_name": ["execute", "executemany", "executescript"],
        # ARG_1 is safe, because the SQL statement is prepared
        # takes no keyword arguments
        "source": "ARG_0",
        "policy_patch": False,
    },
    {
        "module": "mysql.connector",
        "class_name": "Cursor",
        "method_name": ["execute", "executemany"],
        # ARG_1 is safe, because the SQL statement is prepared
        "source": "ARG_0,KWARG:operation",
        "policy_patch": False,
    },
    {
        "module": "mysql.connector.cursor_cext",
        "class_name": "CMySQLCursor",
        "method_name": ["execute", "executemany"],
        # ARG_1 is safe, because the SQL statement is prepared
        "source": "ARG_0,KWARG:operation",
        "policy_patch": False,
    },
    {
        "module": "pymysql",
        "class_name": "Cursor",
        "method_name": ["execute", "executemany"],
        # ARG_1 is safe, because the SQL statement is prepared
        "source": "ARG_0,KWARG:query",
        "policy_patch": False,
    },
    {
        "module": "psycopg2",
        "class_name": "Cursor",
        "method_name": ["execute", "executemany"],
        # ARG_1 is safe, because the SQL statement is prepared
        "source": "ARG_0,KWARG:query",
        "policy_patch": False,
    },
    {
        # Used to handle any SQLAlchemy dialect
        "module": "sqlalchemy",
        "class_name": "Cursor",
        "method_name": ["execute", "executemany"],
        "source": "ARG_0",
        "policy_patch": False,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "sql-injection",
        sql_injection_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .httponly_rule import HttpOnlyRule
from .secure_flag_missing_rule import SecureFlagMissingRule
from .session_timeout_rule import SessionTimeoutRule
from .session_rewriting_rule import SessionRewritingRule

__all__ = [
    "HttpOnlyRule",
    "SecureFlagMissingRule",
    "SessionTimeoutRule",
    "SessionRewritingRule",
]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from zlib import crc32


class Digest:
    def __init__(self):
        self.crc32 = 0

    def finish(self):
        return str(self.crc32)

    def update(self, value: str):
        self.crc32 = crc32(str(value).encode(), self.crc32)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from functools import cached_property

from contrast.agent.protect.rule.mode import Mode


class ProtectRule:
    def __init__(self, protect_rule_json: dict | None):
        self._raw_rule = protect_rule_json or {}

    @cached_property
    def id(self) -> str:
        return self._raw_rule.get("id", "")

    @cached_property
    def mode(self) -> Mode:
        raw_mode = self._raw_rule.get("mode", "OFF")
        raw_block_at_entry = self._raw_rule.get("blockAtEntry", False)

        # Common configuration uses lower case infinitives whereas TeamServer UI uses
        # capital present participles. This mapping has to handle both.
        return {
            "MONITOR": Mode.MONITOR,
            "MONITORING": Mode.MONITOR,
            "BLOCK": Mode.BLOCK,
            "BLOCKING": (Mode.BLOCK_AT_PERIMETER if raw_block_at_entry else Mode.BLOCK),
            "BLOCK_AT_PERIMETER": Mode.BLOCK_AT_PERIMETER,
            "OFF": Mode.OFF,
        }[raw_mode.upper()]



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class Validator:
    def validate(self):
        """
        At this time, any field (optional or required) that does not pass validation
        will cause the entire message to not be reported. In the future we could:

        1. if the field is optional but doesn't pass validation, skip the field
        :return:
        """
        assert self.VALIDATIONS, "Child class must implement VALIDATIONS"

        for field_name, validations in self.VALIDATIONS.items():
            # The field_name we're validating must be defined as an attr to the class.
            try:
                field_value = getattr(self, field_name)
            except AttributeError as e:
                raise ValidationException(
                    self, field_name, "field is not an attr of the class"
                ) from e

            required = validations.get("required")

            # if field is required but it isn't populated.
            if required and not field_value:
                raise ValidationException(
                    self, field_name, "field is required but empty"
                )

            range_val = validations.get("range")

            # python range is exclusive so +1 to make it inclusive of top value
            if range_val is not None and len(field_value) not in range(
                range_val[0], range_val[1] + 1
            ):
                raise ValidationException(
                    self, field_name, "field is outside required range length"
                )


class ValidationException(Exception):
    def __init__(self, klass, field_name, orig_msg):
        message = f"Unable to validate field `{field_name}` for class `{klass.__class__.__name__}`: {orig_msg}"
        super().__init__(message)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import argparse
import os
import sys
import shutil

import contrast
from contrast_rewriter import log_stderr

DESCRIPTION = """
The command-line runner for the Contrast Python Agent.
"""

USAGE = "%(prog)s [-h] -- cmd [cmd ...]"

EPILOG = """
Insert this command before the one you usually use to start your webserver
to apply Contrast's instrumentation. See our public documentation for details:
https://docs.contrastsecurity.com/en/python.html
"""


def _log(msg: str) -> None:
    log_stderr(msg, logger_name="contrast-runner")


def runner() -> None:
    _log("Preparing environment for Contrast instrumentation")
    _log(f"Python Agent version: {contrast.__version__}")
    parser = argparse.ArgumentParser(
        description=DESCRIPTION,
        usage=USAGE,
        epilog=EPILOG,
    )
    # if you add public arguments here, update USAGE accordingly
    parser.add_argument(
        "cmd",
        nargs="+",
        help="Command to run with the agent. cmd should be available in the operating system PATH.",
    )

    parsed = parser.parse_args()

    cmd_path, *args = shutil.which(parsed.cmd[0]), *parsed.cmd
    if cmd_path is None:
        _log(f"ERROR: command not found: {parsed.cmd[0]}")
        _log(f"Run '{sys.argv[0]} --help' for usage information.")
        sys.exit(1)
        return

    loader_path = os.path.join(os.path.dirname(contrast.__file__), "loader")
    os.environ["PYTHONPATH"] = os.path.pathsep.join([loader_path] + sys.path)
    os.environ["CONTRAST_INSTALLATION_TOOL"] = "CONTRAST_PYTHON_RUN"

    _log(f"Entering target process: {parsed.cmd[0]}")
    os.execl(cmd_path, *args)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

# ruff: noqa: F403

from .asgi import *
from .cgi import *
from .django import *
from .falcon import *
from .flask import *
from .quart import *
from .webob import *



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import sys
import time
from logging import INFO

from contrast_vendor import structlog
from contrast_rewriter import LOG_TIME_FORMAT, LOG_FORMAT


def basic_stringifier(logger, method_name, event_dict) -> str:
    date_and_time = time.strftime(LOG_TIME_FORMAT, time.localtime())
    return LOG_FORMAT.format(
        date_and_time=date_and_time,
        logger_name="contrast-agent",
        msg=event_dict.get("event", "<logging error>"),
    )


structlog.configure(
    processors=[basic_stringifier],
    logger_factory=structlog.PrintLoggerFactory(sys.stderr),
    wrapper_class=structlog.make_filtering_bound_logger(INFO),
    cache_logger_on_first_use=False,
)

from typing import TYPE_CHECKING  # noqa: E402

if TYPE_CHECKING:
    from contrast.agent.request_context import RequestContext

import os  # noqa: E402
import contextlib  # noqa: E402
from contextvars import ContextVar  # noqa: E402
from contrast.agent.assess.string_tracker import StringTracker  # noqa: E402
from contrast.version import __version__  # noqa: E402

# aliases
from contrast.agent.assess.utils import get_properties  # noqa: F401 E402

# process globals
REQUEST_CONTEXT: ContextVar[RequestContext | None] = ContextVar(
    "contrast_request_context", default=None
)
STRING_TRACKER = StringTracker()
TELEMETRY = None

# PERF: These values are constant for the lifetime of the agent,
# so we compute them only once instead of potentially computing
# them hundreds of times.
AGENT_CURR_WORKING_DIR = os.getcwd()


def telemetry_disabled() -> bool:
    return os.environ.get("CONTRAST_AGENT_TELEMETRY_OPTOUT", "").lower() in [
        "1",
        "true",
    ]


def get_canonical_version() -> str:
    return ".".join(__version__.split(".")[:3])


class SecurityException(Exception):
    """
    Exception raised by Contrast Protect to block attacks. Full attack details are
    reported to the Contrast UI.
    """

    def __init__(self, *, rule_name: str) -> None:
        super().__init__(
            f"Contrast Protect blocked an attack for rule: {rule_name}. See Contrast UI"
            " for full details."
        )


@contextlib.contextmanager
def lifespan(context: RequestContext):
    # py313 - this method will no longer be necessary when we drop 3.13 because
    # `set()` is a contextmanager in 3.14+
    token = REQUEST_CONTEXT.set(context)
    try:
        yield context
    finally:
        REQUEST_CONTEXT.reset(token)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
import platform
from enum import Enum, EnumMeta
from functools import lru_cache
import itertools
import os
import socket

from contrast import AGENT_CURR_WORKING_DIR
from contrast.utils.string_utils import truncate
from contrast_vendor.ruamel import yaml

from contrast.agent.protect.rule.mode import Mode

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


CONFIG_LOCATIONS = [
    AGENT_CURR_WORKING_DIR,
    os.path.join(AGENT_CURR_WORKING_DIR, "settings"),
    "/etc/contrast/python/",
    "/etc/contrast/",
    "/etc/",
]

CONFIG_FILE_NAMES = ["contrast_security.yaml", "contrast_security.yml"]

DEFAULT_PATHS = [
    os.path.join(path, filename)
    for path, filename in itertools.product(CONFIG_LOCATIONS, CONFIG_FILE_NAMES)
]
"""
DEFAULT_PATHS are the default locations to search for the config file.

Current order of precedence:
    file specified by CONTRAST_CONFIG_PATH env var
    os.getcwd()
    os.path.join(os.getcwd(), 'settings')
    /etc/contrast/python/
    /etc/contrast/
    /etc/
"""

# Valid options are defined in the spec:
# https://github.com/Contrast-Security-Inc/assess-specifications/blob/master/vulnerability/capture-stacktrace.md
STACKTRACE_OPTIONS = ["ALL", "SOME", "NONE", "SINK"]


@lru_cache(maxsize=1)
def get_hostname() -> str:
    """
    Returns the hostname from the socket or "localhost"
    """
    hostname = "localhost"

    try:
        hostname = socket.gethostname() or hostname
    except Exception as e:
        logger.debug(e)

    return truncate(hostname)


@lru_cache(maxsize=1)
def get_platform():
    try:
        platform_str = platform.platform()
    except Exception:
        try:
            platform_str = platform.platform(terse=True)
        except Exception:
            platform_str = "unknown"

    return platform_str


def load_yaml_config(paths: list[str]) -> list[tuple[dict, str]]:
    """
    Loads yaml files at $CONTRAST_CONFIG_PATH or paths as dictionaries.

    See official documentation because this is valid across agents.

    :return: a list of flattened dict object representation of the yaml config. {'enable': True, ....} and filename pairs
    """
    if filename := os.environ.get("CONTRAST_CONFIG_PATH"):
        if os.path.isfile(filename):
            paths = [filename]
        else:
            logger.warning(
                "The path specified by CONTRAST_CONFIG_PATH is not a file -"
                " searching for configuration file in default locations",
                contrast_config_path=filename,
            )

    configs = [
        (flatten_config(cfg), path)
        for path in paths
        if (cfg := _load_config(path)) is not None
    ]

    return configs


def _load_config(file_path):
    try:
        with open(file_path) as config_file:
            logger.info("Loading configuration file: %s", os.path.abspath(file_path))
            return yaml.YAML(typ="safe", pure=True).load(config_file)
    except yaml.scanner.ScannerError as ex:
        # config yaml cannot be loaded but agent should continue on in case
        # env vars are configured
        msg_prefix = "YAML validator found an error."
        msg = f"{msg_prefix} Configuration path: [{ex.problem_mark.name}]. Line [{ex.problem_mark.line}]. Error: {ex.problem}"
        logger.warning(msg)
    except FileNotFoundError:
        # Most paths are system defaults, so we don't want to log a warning.
        # For the user specified CONTRAST_CONFIG_PATH, warnings are logged
        # in load_yaml_config.
        logger.debug("Configuration file not found: %s", file_path)

    return None


def flatten_config(config) -> dict:
    """
    Convert a nested dict such as
        {'enable': True,
        'application':
            {'name': 'dani-flask'},
        'foo':
        'agent':
            {'python':

    into
        'enable': True,
        'application.name': 'dani-flask',

    :param config: dict config with nested keys and values
    :return: dict, flattened where each key has one value.
    """
    flattened_config = {}

    def flatten(x, name=""):
        if isinstance(x, dict):
            for key in x:
                flatten(x[key], name + key + ".")
        elif x is not None:
            flattened_config[name[:-1]] = x

    flatten(config)
    return flattened_config


def str_to_bool(val) -> bool:
    """
    Converts a str to a bool

    true -> True, false -> False
    """
    if isinstance(val, bool):
        return val
    if not isinstance(val, str):
        return False

    # The remainder of this function was ported from distutils
    val = val.lower()

    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    if val in ("n", "no", "f", "false", "off", "0"):
        return False
    raise ValueError(f"invalid truth value {val!r}")


def parse_disabled_rules(disabled_rules) -> list[str]:
    if not disabled_rules:
        return []

    return [rule.lower() for rule in disabled_rules.split(",")]


def parse_stacktraces_options(option) -> str:
    option = option.upper()
    if option in STACKTRACE_OPTIONS:
        return option

    return "ALL"


def parse_event_detail_option(option) -> str:
    return "full" if option.lower() == "full" else "minimal"


def str_to_enum(enum: EnumMeta, *, default: Enum | None = None):
    def _str_to_enum(val: str) -> Enum | None:
        if not val:
            return default
        try:
            return enum[val.upper()]
        except KeyError:
            raise ValueError(f"Invalid value for {enum.__name__}: {val}") from None

    return _str_to_enum


def str_to_protect_mode_enum(mode) -> Mode:
    """
    Converts str config value to protect mode enum that the agent understands
    """
    if not mode:
        return Mode.OFF

    return getattr(Mode, mode.upper(), Mode.MONITOR)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from __future__ import annotations
from pathlib import Path

import functools

from copy import copy
from importlib import import_module
from types import FunctionType
from contrast_fireball import DiscoveredRoute
from django.urls import ResolverMatch, get_resolver
from django.urls.exceptions import Resolver404
from django.urls.resolvers import (
    URLPattern as RegexURLPattern,
    URLResolver as RegexURLResolver,
)

from contrast.agent.middlewares.route_coverage.common import (
    DEFAULT_ROUTE_METHODS,
    build_args_from_function,
)

from contrast_vendor import structlog as logging
from contrast.utils.decorators import fail_quietly

logger = logging.getLogger("contrast")


def get_required_http_methods(func: FunctionType) -> set | None:
    """
    Grabs the require_http_list closure variable from a view function through its code object.
    """

    if isinstance(func, functools.partial):
        func = getattr(func, "func", None)

    if func is None:
        return None

    method_types = _get_required_http_methods(func)

    if (wrapped := getattr(func, "__wrapped__", None)) and (
        restricted_methods := get_required_http_methods(wrapped)
    ) is not None:
        method_types = (
            set(restricted_methods)
            if method_types is None
            else method_types.intersection(restricted_methods)
        )

    return method_types


def _get_closure_variable(func: FunctionType, varname: str):
    if not (closure := getattr(func, "__closure__", None)):
        return None
    if varname not in func.__code__.co_freevars:
        return None
    index = func.__code__.co_freevars.index(varname)
    return closure[index].cell_contents


DJANGO_HTTP_DECORATOR_PATH_SUFFIX = str(
    Path("django", "views", "decorators", "http.py")
)


def _get_required_http_methods(viewfunc: FunctionType):
    if class_view := getattr(viewfunc, "view_class", None):
        # This is a class-based view,
        return {
            method.upper()
            for method in class_view.http_method_names
            if hasattr(class_view, method)
        }
    if not viewfunc.__code__.co_filename.endswith(DJANGO_HTTP_DECORATOR_PATH_SUFFIX):
        return None
    restricted_methods = _get_closure_variable(viewfunc, "request_method_list")
    if restricted_methods and isinstance(restricted_methods, list):
        return set(restricted_methods)
    return None


def get_method_info(pattern_or_resolver):
    if not (viewfunc := pattern_or_resolver.callback):
        return DEFAULT_ROUTE_METHODS, "()"

    method_arg_names = build_args_from_function(viewfunc)
    method_types = (
        required_methods
        if (required_methods := get_required_http_methods(viewfunc)) is not None
        else DEFAULT_ROUTE_METHODS
    )
    return method_types, method_arg_names


def create_routes(urlpatterns, url_prefix="/") -> set[DiscoveredRoute]:
    routes = set()

    def url_str(pattern: str):
        return f"{url_prefix.removesuffix('/')}/{pattern}"

    for urlpattern in reversed(urlpatterns):
        if isinstance(urlpattern, RegexURLResolver):
            routes.update(
                create_routes(urlpattern.url_patterns, url_str(str(urlpattern.pattern)))
            )

        elif isinstance(urlpattern, RegexURLPattern):
            method_types, method_arg_names = get_method_info(urlpattern)
            url = url_str(str(urlpattern.pattern))
            signature = build_django_signature(urlpattern, method_arg_names)
            for method_type in method_types:
                routes.add(
                    DiscoveredRoute(
                        verb=method_type,
                        url=url,
                        signature=signature,
                        framework="Django",
                    )
                )

    return routes


def create_django_routes() -> set[DiscoveredRoute]:
    """
    Grabs all URL's from the root settings and searches for possible required_method decorators

    In Django there is no implicit declaration of GET or POST. Often times decorators are used to fix this.

    Returns a dict of key = id, value = api.Route.
    """

    from django.conf import settings

    if not settings.ROOT_URLCONF:
        logger.info("Application does not define settings.ROOT_URLCONF")
        logger.debug("Skipping enumeration of urlpatterns")
        return set()

    try:
        root_urlconf = import_module(settings.ROOT_URLCONF)
    except Exception as exception:
        logger.debug("Failed to import ROOT_URLCONF: %s", exception)
        return set()

    try:
        urlpatterns = root_urlconf.urlpatterns or []
    except Exception as exception:
        logger.debug("Failed to get urlpatterns: %s", exception)
        return set()

    url_patterns = copy(urlpatterns)
    return create_routes(url_patterns)


def _function_loc(func):
    """Return the function's module and name"""
    return f"{func.__module__}.{func.__name__}"


def build_django_signature(obj, method_arg_names=None):
    if hasattr(obj, "lookup_str"):
        signature = obj.lookup_str
    elif hasattr(obj, "_func_path"):
        signature = obj._func_path
        obj = obj.func
    elif hasattr(obj, "callback"):
        cb = obj.callback
        signature = _function_loc(cb)
    elif callable(obj):
        signature = _function_loc(obj)
    else:
        logger.debug(
            "WARNING: can't build django signature for object type %s", type(obj)
        )
        return ""

    if method_arg_names is None:
        method_arg_names = build_args_from_function(obj)

    signature += method_arg_names
    return signature


@fail_quietly("Failed to get view function for django application")
def get_matched_resolver(path) -> ResolverMatch | None:
    from django.conf import settings

    try:
        result = get_resolver().resolve(path or "/")
    except Resolver404:
        return None

    if (
        result is None
        and not path.endswith("/")
        and "django.middleware.common.CommonMiddleware" in settings.MIDDLEWARE
        and settings.APPEND_SLASH
    ):
        result = get_matched_resolver(f"{path}/")
    if result is None:
        return None

    return result



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.utils.configuration_utils import str_to_bool
from .config_builder import ConfigBuilder
from .config_option import ConfigOption
from contrast import AGENT_CURR_WORKING_DIR
import os


class Application(ConfigBuilder):
    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="application.code", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.group", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.sensitive_data_masking_policy.mask_attack_vector",
                default_value=False,
                type_cast=str_to_bool,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="application.sensitive_data_masking_policy.mask_http_body",
                default_value=False,
                type_cast=str_to_bool,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="application.sensitive_data_masking_policy.rules",
                default_value=[],
                type_cast=list,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="application.input_exclusions",
                default_value=[],
                type_cast=list,
                log_effective_config=False,
            ),
            ConfigOption(
                canonical_name="application.metadata", default_value="", type_cast=str
            ),
            ConfigOption(
                # TODO: PYT-2852 Revisit application name detection
                canonical_name="application.name",
                default_value=os.path.basename(AGENT_CURR_WORKING_DIR),
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="application.path", default_value="/", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.tags", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.version", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.session_id", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="application.session_metadata",
                default_value="",
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="application.url_exclusions",
                default_value=[],
                type_cast=list,
                log_effective_config=False,
            ),
        ]



from __future__ import annotations

import httpx
import os
import subprocess

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from .models import Message
import json


mock_httpx_transport = httpx.MockTransport(
    lambda _: httpx.Response(200, text="mocked response for testing")
)


def shell_clock(request: HttpRequest):
    format = request.GET.get("format", "+%H:%M:%S")
    time = subprocess.check_output(f"date {format}", shell=True).strip().decode()
    return HttpResponse(
        f"<html><body><h1>Current Time</h1><p>{time}</p></body></html>".encode()
    )


def system_sleep(request: HttpRequest):
    duration = float(request.GET.get("duration", 0.1))
    os.system(f"sleep {duration}")
    return HttpResponse(
        f"<html><body><h1>Slept for {duration} seconds</h1></body></html>".encode()
    )


@method_decorator(csrf_exempt, name="dispatch")
class PostsView(View):
    POSTS_DIR = "posts"

    def get(self, request: HttpRequest, name: str | None = None):
        if name is None:
            # List all posts
            posts = os.listdir(self.POSTS_DIR)
            posts_list = "".join(f"<li>{post}</li>" for post in posts)
            return HttpResponse(
                f"<html><body><h1>Posts</h1><ul>{posts_list}</ul></body></html>".encode()
            )
        else:
            post_path = os.path.join(self.POSTS_DIR, name)
            if not os.path.isfile(post_path):
                return HttpResponse(
                    f"<html><body><h1>Not Found</h1><p>Post '{name}' does not exist.</p></body></html>".encode(),
                    status=404,
                )
            with open(post_path, "rb") as f:
                content = f.read()
            return HttpResponse(content)

    def post(self, request: HttpRequest, name: str):
        post_path = os.path.join(self.POSTS_DIR, name)
        with open(post_path, "wb") as f:
            f.write(request.body)
        return HttpResponse(
            f"<html><body><h1>Post '{name}' saved.</h1></body></html>".encode()
        )


# Prep the filesystem for PostsView
os.makedirs(PostsView.POSTS_DIR, exist_ok=True)
example_post_path = os.path.join(PostsView.POSTS_DIR, "example.txt")
with open(example_post_path, "w") as f:
    f.write("This is an example post.\n")


@csrf_exempt
def messages_view(request: HttpRequest):
    """
    GET: List all messages as JSON.
    POST: Create a new message with JSON body: {"text": "..."}
    """
    if request.method == "GET":
        messages = Message.objects.order_by("-created_at")
        data = [
            {"id": m.id, "text": m.text, "created_at": m.created_at.isoformat()}
            for m in messages
        ]
        return JsonResponse(data, safe=False)
    elif request.method == "POST":
        try:
            payload = json.loads(request.body.decode())
            text = payload.get("text", "")
            if not text:
                return JsonResponse({"error": "Missing 'text' field"}, status=400)
            msg = Message.objects.create(text=text)
            return JsonResponse(
                {
                    "id": msg.id,
                    "text": msg.text,
                    "created_at": msg.created_at.isoformat(),
                },
                status=201,
            )
        except Exception as e:
            return JsonResponse({"error": str(e)}, status=400)
    else:
        return JsonResponse({"error": "Method not allowed"}, status=405)


def authentication_test(request):
    if request.user.is_authenticated:
        return HttpResponse(
            f"<html><body><h1>Authenticated</h1><p>User: {request.user.username}</p></body></html>".encode()
        )
    else:
        return HttpResponse(
            b"<html><body><h1>Not Authenticated</h1></body></html>",
            status=401,
        )


def external_data(request: HttpRequest):
    # does not send a real http request (for testing purposes)
    with httpx.Client(transport=mock_httpx_transport) as client:
        response = client.get(
            "http://user1:passw0rd@externalservice.com:8080/fetch/?q=all"
        )
    if response.status_code == 200:
        return JsonResponse({"result": response.text})
    else:
        return JsonResponse({"error": "Failed to fetch data"}, status=503)



# Copyright © 2025 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from dataclasses import dataclass, replace
from typing import TypedDict
from urllib.parse import parse_qs, unquote, urlencode
from collections.abc import Mapping

from contrast.api.attack import ProtectResponse
from contrast.api.user_input import UserInput
from contrast_vendor import structlog as logging
from contrast_vendor.webob.multidict import MultiDict

logger = logging.getLogger("contrast")

MASK = "contrast-redacted-{}"
VECTOR_MASK = MASK.format("vector")
BODY_MASK = "contrast-redacted-body"
SEMICOLON_URL_ENCODE_VAL = "%25"


class SensitiveDataRule(TypedDict):
    id: str
    keywords: list[str]


@dataclass(frozen=True)
class SensitiveDataPolicy:
    mask_attack_vector: bool
    mask_http_body: bool
    rules: list[SensitiveDataRule]


class RequestMasker:
    def __init__(self, config: Mapping):
        self.mask_rules = SensitiveDataPolicy(
            mask_attack_vector=config[
                "application.sensitive_data_masking_policy.mask_attack_vector"
            ],
            mask_http_body=config[
                "application.sensitive_data_masking_policy.mask_http_body"
            ],
            rules=config["application.sensitive_data_masking_policy.rules"],
        )

    @classmethod
    def new_request_masker(cls, config: SensitiveDataPolicy):
        return cls(
            {
                "application.sensitive_data_masking_policy.mask_attack_vector": config.mask_attack_vector,
                "application.sensitive_data_masking_policy.mask_http_body": config.mask_http_body,
                "application.sensitive_data_masking_policy.rules": config.rules,
            }
        )

    def mask_attack_input(self, input: UserInput):
        """
        Mask the attack input based on the masking rules.
        """
        masked_name = (
            self.mask_attack_vector(input.name)
            if input.is_name_based and input.name
            else input.name
        )

        return replace(
            input,
            value=self.mask_attack_vector(input.value),
            name=masked_name,
        )

    def mask_attack_vector(self, vector: str):
        """
        Mask the provided vector if "mask_attack_vector" is set to True in the mask rules.
        """
        return vector if not self.mask_rules.mask_attack_vector else VECTOR_MASK

    def mask_sensitive_data(self, request, attack=None):
        if not request or not self.mask_rules:
            return

        self.request = request
        self.attack = attack

        logger.debug("Masker: masking sensitive data")

        self._mask_body()
        self._mask_query_string()
        self._mask_request_params()
        self._mask_request_cookies()
        self._mask_request_headers()

        request._masked = True

    def _mask_body(self):
        if not self.request.body:
            return

        if self.mask_rules.mask_http_body:
            self.request._masked_body = BODY_MASK
            return
        if (
            self.mask_rules.mask_attack_vector
            and self.attack
            and any(
                sample.user_input.type.is_body_based for sample in self.attack.samples
            )
        ):
            # NOTE: is_body_based is loose when it comes to querystring parameters.
            # This could cause extra redaction, but that's an acceptable trade-off
            # in the short term.
            self.request._masked_body = VECTOR_MASK

    def _mask_query_string(self):
        if self.request.query_string:
            self.request._masked_query_string = self._mask_raw_query(
                self.request.query_string
            )

    def _mask_raw_query(self, query_string):
        qs_dict = parse_qs(query_string)
        masked_qs_dict = self._mask_dictionary(qs_dict)
        return urlencode(masked_qs_dict, doseq=True)

    def _mask_request_params(self):
        params = self.request.params
        if not params:
            return

        self.request._masked_params = self._mask_dictionary(params)

    def _mask_request_cookies(self):
        cookies = self.request.cookies
        if not cookies:
            return

        self.request._masked_cookies = self._mask_dictionary(cookies)

    def _mask_request_headers(self):
        headers = self.request.headers
        if not headers:
            return

        self.request._masked_headers = self._mask_dictionary(headers)

    def _mask_dictionary(self, d):
        if not d:
            return d

        if isinstance(d, MultiDict):
            d = d.mixed()

        return {
            self._mask_key_vector(k, self.attack): self._mask_value(k, v, self.attack)
            for k, v in d.items()
        }

    def _mask_key_vector(self, k, attack):
        return (
            VECTOR_MASK
            if self.mask_rules.mask_attack_vector and self._is_value_vector(attack, k)
            else k
        )

    def _mask_value(self, k, v, attack):
        if isinstance(v, list):
            return [self._mask_value(k, item, attack) for item in v]

        if not self._is_value_vector(attack, v):
            if k is not None and self._find_value_index_in_rules(k.lower()) != -1:
                return MASK.format(k.lower())
        elif self.mask_rules.mask_attack_vector:
            return VECTOR_MASK
        return v

    def _is_value_vector(self, attack, value):
        if not attack or attack.response == ProtectResponse.NO_ACTION or not value:
            return False

        return self._is_value_in_sample(attack.samples, value)

    def _is_value_in_sample(self, samples, value):
        if not samples:
            return False

        # Setting this to remove url encoding of header and cookie values
        value = unquote(value)

        return any(sample.user_input.value == value for sample in samples)

    def _find_value_index_in_rules(self, s):
        index = -1
        # When looking for header it replaces '_' with '-' and I don't want to risk not
        # properly matching to the rules
        s = s.replace("-", "_")
        for rule in self.mask_rules.rules:
            try:
                index = rule.get("keywords").index(s)
                break
            except ValueError:
                index = -1

        return index
